import grpc
from concurrent import futures

import greeter_pb2
import greeter_pb2_grpc


class GreeterServicer(greeter_pb2_grpc.GreeterServicer):

    def Echo(self, request, context):
        response = greeter_pb2.HelloResponse()
        response.name = request.name
        return response

    def Greet(self, request, context):
        response = greeter_pb2.HelloResponse()
        response.name = 'Result for ' + request.result.title + ' contains ' + request.result.snippets
        return response


server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
greeter_pb2_grpc.add_GreeterServicer_to_server(GreeterServicer(), server)
server.add_insecure_port('[::]:50051')



import datetime

from catcher.utils.misc import try_get_object, fill_template

from test.abs_test_class import TestClass


class MiscTest(TestClass):
    def __init__(self, method_name):
        super().__init__('misc_test', method_name)

    # read and parse yaml file
    def test_try_get_object(self):
        not_an_object = 'not an not_an_object'
        self.assertEqual(not_an_object, try_get_object(not_an_object))

        python_term = '{\'key\': \'string\', \'int_key\': 1, \'sub_term\': {\'k\':[1,2,3]}, \'non_existent\':None}'
        python_object = {'key': 'string', 'int_key': 1, 'sub_term': {'k': [1, 2, 3]}, 'non_existent': None}
        self.assertEqual(python_object, try_get_object(python_term))

        json_object = '{"key": "string", "int_key": 1, "sub_term": {"k":[1,2,3]}, "non_existent":null}'
        self.assertEqual(python_object, try_get_object(json_object))

    # date is parsed without problems
    def test_try_get_date(self):
        data = datetime.datetime(2020, 3, 20, 17, 56, 34, 74807)
        self.assertEqual(data, try_get_object(data.__repr__()))

        obj_str = "{'key':'value','date':" + data.__repr__() + '}'
        self.assertEqual({'key': 'value', 'date': data}, try_get_object(obj_str))

    # primitives are parsed and returned
    def test_try_get_primitives(self):
        self.assertEqual(True, try_get_object('True'))
        self.assertEqual(True, try_get_object('true'))
        self.assertEqual(11, try_get_object('11'))
        self.assertEqual(7.3, try_get_object('7.3'))
        self.assertEqual('string', try_get_object('string'))
        self.assertEqual('id', try_get_object('id'))
        self.assertEqual('2020-03-20', try_get_object('2020-03-20'))

    # complex objects are parsed and returned
    def test_try_json(self):
        my_dict = {'key': 'value', 'inner': {'key': 'value'}}
        self.assertEqual(my_dict, try_get_object('{"key":"value","inner":{"key":"value"}}'))
        my_list = [1, 2, 3, [1, 2, 3]]
        self.assertEqual(my_list, try_get_object('[1, 2, 3, [1, 2, 3]]'))

    # date with compound object should be rendered as a string
    def test_render_date_nested(self):
        now = datetime.datetime.now()
        res = fill_template('{{ OUTPUT }}', {'OUTPUT': {'key': 'value1', 'date': now}})
        self.assertEqual(res['date'], now.strftime('%Y-%m-%d %H:%M:%S.%f'))
        self.assertEqual(res['key'], 'value1')

    # templates are being filled in correctly with primitives
    def test_template_with_simple_object(self):
        self.assertEqual(True, fill_template('{{ OUTPUT }}', {'OUTPUT': True}))
        self.assertEqual(17, fill_template('{{ OUTPUT }}', {'OUTPUT': 17}))
        self.assertEqual(4.2, fill_template('{{ OUTPUT }}', {'OUTPUT': 4.2}))
        self.assertEqual('string', fill_template('{{ OUTPUT }}', {'OUTPUT': 'string'}))
        self.assertEqual('id', fill_template('{{ OUTPUT }}', {'OUTPUT': 'id'}))
        self.assertEqual('2020-03-11', fill_template('{{ OUTPUT }}', {'OUTPUT': '2020-03-11'}))

    # templates are being filled in correctly with complex objects
    def test_template_with_complex_object(self):
        my_dict = {'key': 'value', 'inner': {'key': 'value'}}
        self.assertEqual(my_dict, fill_template('{{ OUTPUT }}', {'OUTPUT': my_dict}))
        my_list = [1, 2, 3, [1, 2, 3]]
        self.assertEqual(my_list, fill_template('{{ OUTPUT }}', {'OUTPUT': my_list}))
        my_tuple = (1, 2, (1, 2))
        self.assertEqual(my_tuple, fill_template('{{ OUTPUT }}', {'OUTPUT': my_tuple}))
        my_complex = [(1, {'foo': [1, 2, 3]}, [1, 2, {1: 'a'}])]
        self.assertEqual(my_complex, fill_template('{{ OUTPUT }}', {'OUTPUT': my_complex}))



import time
from abc import ABC, abstractmethod
from os.path import join
from typing import Union


class Formatter(ABC):
    """
    Formatter process logs in log storage and saves it in the desired output
    """

    @abstractmethod
    def format(self, path: str, data: list):
        pass


class JsonFormatter(Formatter):
    def format(self, path: str, data: list):
        import json
        with open(join(path, 'report_' + str(time.time()) + '.json'), 'w') as fp:
            json.dump(data, fp)


class HTMLFormatter(Formatter):
    def format(self, path: str, data: list):
        pass


class OutputFormatter(Formatter):
    def format(self, path: str, data: list):
        print(data)


def formatter_factory(out_format: str) -> Union[Formatter, None]:
    out_format = out_format.lower()
    if out_format == 'json':
        return JsonFormatter()
    elif out_format == 'html':
        return HTMLFormatter()
    else:
        return OutputFormatter()



import ipaddress
import random
from os.path import join

import pytest
from faker import Faker

from test.abs_test_class import TestClass
from catcher.core.runner import Runner
from test.test_utils import check_file, read_file
from catcher.utils.singleton import Singleton


class FiltersTest(TestClass):
    def __init__(self, method_name):
        super().__init__('filters_test', method_name)
        Singleton._instances = {}

    def test_builtin_filters_available(self):
        self.populate_file('main.yaml', '''---
                variables:
                    my_var: test
                steps:
                    - echo: {from: '{{ "test" | hash }}', to: one.output}
                    - echo: {from: '{{ my_var | hash("sha1") }}', to: two.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), '098f6bcd4621d373cade4e832627b4f6'))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), 'a94a8fe5ccb19ba61c4c0873d391e987982fbbd3'))

    def test_builtin_functions_available(self):
        self.populate_file('main.yaml', '''---
                variables:
                    my_list: ['one', 'two', 'three']
                steps:
                    - echo: {from: '{{ random("ipv4_private") }}', to: one.output}
                    - echo: {from: '{{ random_int(1, 10) }}', to: two.output}
                    - echo: {from: '{{ random_choice(my_list) }}', to: three.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        Faker.seed(4321)
        random.seed(123)
        self.assertTrue(runner.run_tests())
        ipaddress.ip_address(read_file(join(self.test_dir, 'one.output')))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), '1'))
        self.assertTrue(check_file(join(self.test_dir, 'three.output'), 'three'))

    def test_custom_filters_available(self):
        self.populate_file('custom_filter.py',
                           '''def filter_increment(input):
       if isinstance(input, int):
         return input + 1
       return 'not an int'
    
def filter_to_string(input, arg='test'):
  return _not_a_fun(input) + ' ' + arg

def _not_a_fun(arg):
  return 'to_str=' + str(arg)  
                            ''')
        self.populate_file('main.yaml', '''---
                        steps:
                            - echo: {from: '{{ 221 | increment }}', to: one.output}
                            - echo: {from: '{{ 221 | to_string("hello") }}', to: two.output}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None,
                        filter_list=[join(self.test_dir, 'custom_filter.py')])
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), '222'))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), 'to_str=221 hello'))

    def test_custom_functions_available(self):
        self.populate_file('custom_filter.py',
                           '''def function_my_custom():
                               return {'key': 'value'}

def function_my_other(input):
  return _not_a_fun(input)

def _not_a_fun(arg):
  return 'to_str=' + str(arg)  
                            ''')
        self.populate_file('main.yaml', '''---
                                steps:
                                    - echo: {from: '{{ my_custom() }}', to: one.output}
                                    - echo: {from: '{{ my_other("test") }}', to: two.output}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None,
                        filter_list=[join(self.test_dir, 'custom_filter.py')])
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), "{'key': 'value'}"))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), 'to_str=test'))

    def test_custom_both_available(self):
        self.populate_file('custom_filter.py',
                           '''def filter_increment(input):
       if isinstance(input, int):
         return input + 1
       return 'not an int'
       
def function_my_other(input):
  return _not_a_fun(input)

def _not_a_fun(arg):
  return 'to_str=' + str(arg)  
                            ''')
        self.populate_file('main.yaml', '''---
                                steps:
                                    - echo: {from: '{{ 221 | increment }}', to: one.output}
                                    - echo: {from: '{{ my_other("test") }}', to: two.output}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None,
                        filter_list=[join(self.test_dir, 'custom_filter.py')])
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), '222'))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), 'to_str=test'))

    def test_custom_ignore_other(self):
        self.populate_file('custom_filter.py',
                           '''def other(arg):
  return 'to_str=' + str(arg)  
                            ''')
        self.populate_file('main.yaml', '''---
                                steps:
                                    - echo: {from: '{{ 221 | other }}', to: one.output}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None,
                        filter_list=[join(self.test_dir, 'custom_filter.py')])
        self.assertFalse(runner.run_tests())

    @pytest.mark.skip(reason="can't catch bug with timezones. Disabled for CI.")
    def test_date_time_filters(self):
        self.populate_file('main.yaml', '''---
                                variables:
                                    date_time: '2020-03-30 11:21:47.455790'
                                    timestamp: 1585560107.45579
                                steps:
                                    - echo: {from: '{{ date_time | astimestamp }}', to: one.output}
                                    - echo: {from: '{{ timestamp | asdate }}', to: two.output}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), "1585560107.45579"))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), '2020-03-30 11:21:47.455790'))

    def test_dt_filters_compatible(self):
        """
        old (variable based) and new (function based) formats are not compatible, as variable based date format
        drops milliseconds. To make this test pass I have to add current_ts.split(".")[0]+".0" which also drops it
        """
        self.populate_file('main.yaml', '''---
                            variables:
                                current_dt: '{{ NOW_DT }}'
                                current_ts: '{{ NOW_TS }}'
                            steps:
                                - check: {equals: {the: '{{ current_dt }}', 
                                is: '{{ current_ts | asdate(date_format="%Y-%m-%dT%H:%M:%S0+0000") }}'}}
                                - check: {equals: {the: '{{ current_ts.split(".")[0]+".0" }}', 
                                is: '{{ current_dt | astimestamp(date_format="%Y-%m-%dT%H:%M:%S0+0000") }}'}}
                            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_dt_fitlers_revertable(self):
        self.populate_file('main.yaml', '''---
                                variables:
                                    current_ts: '{{ now_ts() }}'
                                steps:
                                    - echo: {from: '{{ current_ts | asdate }}', register: {current_dt: '{{ OUTPUT }}'}}
                                    - check: {equals: {the: '{{ current_ts }}', is: '{{ current_dt | astimestamp }}'}}
                                    - check: {equals: {the: '{{ current_dt }}', is: '{{ current_ts | asdate }}'}}
                                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_data_conversion_filters(self):
        self.populate_file('main.yaml', '''---
                        variables:
                            my_tuple: ('one', 'two', 'three')
                            my_list: ['one', 'two', 'three']
                            my_dict: {'one': 1, 'two': 2}
                            my_int: 17
                            my_float: 36.6
                        steps:
                            - check: {equals: {the: '{{ my_tuple }}', is: '{{ my_list | astuple }}'}}
                            - check: {equals: {the: '{{ my_list }}', is: '{{ my_tuple | aslist }}'}}
                            - check: {equals: {the: '{{ my_int }}', is: '{{ "17" | asint }}'}}
                            - check: {equals: {the: '{{ my_float }}', is: '{{ "36.6" | asfloat }}'}}
                            - check: {equals: {the: '{{ my_dict }}', is: '{{ [("one", 1), ("two", 2)] | asdict }}'}}
                            - check: {equals: {the: [2, 2], 
                                                is: '{{ ([("a", 2), ("b", 2)] | asdict).values() |aslist }}'}}
                            - check: {equals: {the: '17', is: '{{ my_int | asstr }}'}}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # hash function is properly called
    def test_hash(self):
        self.populate_file('main.yaml', '''---
        variables:
            my_var: test
        steps:
            - echo: {from: '{{ "test" | hash("sha1") }}', to: one.output}
            - echo: {from: '{{ my_var | hash("sha1") }}', to: two.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'a94a8fe5ccb19ba61c4c0873d391e987982fbbd3'))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), 'a94a8fe5ccb19ba61c4c0873d391e987982fbbd3'))

    # random choice on a list can be called
    def test_random_choice(self):
        self.populate_file('main.yaml', '''---
        variables:
            my_list: ['one', 'two', 'three']
        steps:
            - echo: {from: '{{ random_choice(my_list) }}', to: one.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        random.seed(123)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'three'))

    # random int with args can be called
    def test_random_int(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: '{{ random_int(1, 10) }}', to: one.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        random.seed(123)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), '9'))

        # no upper limit
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: '{{ random_int(1) }}', to: one.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        random.seed(123)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), '5186474716495645053'))

        # no lower limit
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: '{{ random_int(range_to=1) }}', to: one.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        random.seed(123)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), '-2936754363331581815'))

    # faker can be called from catcher
    def test_random_functions(self):
        Faker.seed(4321)
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: '{{ random("ipv4_private") }}', to: one.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        ipaddress.ip_address(read_file(join(self.test_dir, 'one.output')))

    def test_base64_filters(self):
        self.populate_file('main.yaml', '''---
                variables:
                    foo: 'foo:bar'
                steps:
                    - echo: {from: '{{ foo |b64encode }}', register: {encoded: '{{ OUTPUT }}'}}
                    - check: 
                        equals: {the: '{{ encoded }}', is: 'Zm9vOmJhcg=='}
                    - echo: {from: '{{ encoded |b64decode }}', register: {decoded: '{{ OUTPUT }}'}}
                    - check: 
                        equals: {the: '{{ decoded }}', is: '{{ foo }}'}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())



import ast
import os
from os.path import join

from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from test.test_utils import check_file, read_file


class InventoryTest(TestClass):
    def __init__(self, method_name):
        super().__init__('inventory_test', method_name)

    # test inventory vars available in test
    def test_inventory_vars_available(self):
        self.populate_file('inventory.yml', '''---
        prod_url: http://prod.url
        foo: bar
        ''')
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: '{{ prod_url }}', to: main1.output}
            - echo: {from: '{{ foo }}', to: main2.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), join(self.test_dir, 'inventory.yml'))
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'main1.output'), 'http://prod.url'))
        self.assertTrue(check_file(join(self.test_dir, 'main2.output'), 'bar'))

    # test inventory vars available in include
    def test_inventory_vars_available_in_include(self):
        self.populate_file('inventory.yml', '''---
        foo: bar
        ''')
        self.populate_file('main.yaml', '''---
        include: one.yaml
        steps:
            - echo: {from: '{{ foo }}', to: main.output}
        ''')
        self.populate_file('one.yaml', '''---
        steps:
            - echo: {from: '{{ foo }}', to: include.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), join(self.test_dir, 'inventory.yml'))
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'include.output'), 'bar'))

    # test inventory vars can be overridden with test vars
    def test_override_inventory(self):
        self.populate_file('inventory.yml', '''---
        foo: bar
        ''')
        self.populate_file('main.yaml', '''---
        variables:
            foo: baz
        steps:
            - echo: {from: '{{ foo }}', to: main.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), join(self.test_dir, 'inventory.yml'))
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'main.output'), 'baz'))

    def test_cmd_variable_override_inventory(self):
        self.populate_file('inventory.yml', '''---
        one: bar
        ''')
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: '{{ one }}', to: 1.output}
            - echo: {from: '{{ two }}', to: 2.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), join(self.test_dir, 'inventory.yml'),
                        cmd_env={'one': 'baz', 'two': 'bax'})
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, '1.output'), 'baz'))
        self.assertTrue(check_file(join(self.test_dir, '2.output'), 'bax'))

    # env var can be set in vars
    def test_env_var_in_inventory(self):
        os.environ['FOO'] = '1'
        self.populate_file('inventory.yml', '''---
        foo: '{{ FOO }}'
        ''')
        self.populate_file('main.yaml', '''---
        steps:
            - check: {equals: {the: '{{ foo }}', is: '1'}}
        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        join(self.test_dir, 'inventory.yml'),
                        system_environment=dict(os.environ))
        self.assertTrue(runner.run_tests())

    # env var can be set inside other vars
    def test_env_vars_complex(self):
        os.environ['USER'] = 'user'
        os.environ['PASSWORD'] = '********'
        self.populate_file('inventory.yml', '''---
        database_conf:
            username: '{{ USER }}'
            password: '{{ PASSWORD }}'
        ''')
        self.populate_file('main.yaml', '''---
        steps:
           - echo: {from: '{{ database_conf }}', to: database_conf.output}
           - check: {equals: {the: '{{ database_conf.username }}', is: 'user'}}
        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        join(self.test_dir, 'inventory.yml'),
                        system_environment=dict(os.environ))
        self.assertTrue(runner.run_tests())
        content = ast.literal_eval(read_file(join(self.test_dir, 'database_conf.output')))
        self.assertEquals(content, {'username': 'user', 'password': '********'})

    def test_ignore_locally(self):
        self.populate_file('local.yml', '''---
                service:
                    host: '127.0.0.1'
                    port: '8080'
                ''')
        self.populate_file('cloud.yml', '''---
                service:
                    host: 'my.gateway/service'
                    port: '9000'
                ''')
        self.populate_file('main.yaml', '''---
                                ignore: 
                                    equals: {the: '{{ INVENTORY }}', is: 'local'}
                                steps:
                                    - check: {equals: {the: true, is: false}}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), join(self.test_dir, 'local.yml'))
        self.assertTrue(runner.run_tests())

        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), join(self.test_dir, 'cloud.yml'))
        self.assertFalse(runner.run_tests())



from os.path import join

from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from test.test_utils import check_file


class LoopTest(TestClass):
    def __init__(self, method_name):
        super().__init__('loop_test', method_name)

    def test_while_short(self):
        self.populate_file('main.yaml', '''---
        variables:
            counter: 0
        steps:
            - loop: 
                while:
                    if: '{{ counter < 10 }}'
                    do:
                        echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}'}}
            - check: '{{ counter == 10 }}'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_while_full(self):
        self.populate_file('main.yaml', '''---
        variables:
            counter: 0
        steps:
            - loop: 
                while:
                    if: 
                        equals: {the: '{{ counter }}', is_not: 10}
                    do:
                        echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}'}}
            - check: '{{ counter == 10 }}'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test if multiple actions is available and variables from previous actions is available in current one
    def test_while_multiple_actions(self):
        self.populate_file('main.yaml', '''---
        variables:
            counter: 0
        steps:
            - loop: 
                while:
                    if: '{{ counter < 10 }}'
                    do:
                        - echo: {from: '{{ result }}', to: '{{ counter }}.output'}
                        - echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}', result: '{{ OUTPUT }}'}}
            - check: '{{ counter == 10 }}'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, '0.output'), ''))
        for i in range(1, 10):
            self.assertTrue(check_file(join(self.test_dir, str(i) + '.output'), str(i)))

    def test_max_cycle(self):
        self.populate_file('main.yaml', '''---
        variables:
            counter: 0
        steps:
            - loop: 
                while:
                    if: 
                        equals: {the: '{{ counter }}', is_not: 10000}
                    do:
                        echo: {from: '{{ counter + 1}}', register: {counter: '{{ OUTPUT }}'}}
                    max_cycle: 10
            - check: '{{ counter == 11 }}'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_for(self):
        self.populate_file('main.yaml', '''---
        variables:
            list: [1, 2, 3, 4, 5, 6]
        steps:
            - loop: 
                foreach:
                    in: '{{ list }}'
                    do:
                        echo: {from: '{{ ITEM }}', to: '{{ ITEM }}.output'}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        for i in range(1, 6):
            self.assertTrue(check_file(join(self.test_dir, str(i) + '.output'), str(i)))

    # check multiple actions are allowed, last statement's return is available.
    def test_for_multiple_actions(self):
        self.populate_file('main.yaml', '''---
        variables:
            list: [1, 2, 3, 4, 5, 6]
        steps:
            - loop: 
                foreach:
                    in: '{{ list }}'
                    do:
                        - echo: {from: '{{ ITEM }}', register: {output: '{{ OUTPUT }}'}}
                        - echo: {from: '{{ ITEM }}', to: '{{ ITEM }}.output'}
            - check: '{{ output == 6 }}'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_iterate_dict(self):
        self.populate_file('main.yaml', '''---
        variables:
            list: [{'key': 'a', 'value': 1}, {'key': 'b', 'value': 2}, {'key': 'c', 'value': 3}]
        steps:
            - loop: 
                foreach:
                    in: '{{ list }}'
                    do:
                        - echo: {from: '{{ ITEM }}', register: {output: '{{ OUTPUT }}'}}
                        - echo: {from: '{{ ITEM["value"] }}', to: '{{ ITEM["key"] }}.output'}
            - check: '{{ output == {"key": "c", "value": 3} }}'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'a.output'), '1'))
        self.assertTrue(check_file(join(self.test_dir, 'b.output'), '2'))
        self.assertTrue(check_file(join(self.test_dir, 'c.output'), '3'))

    def test_loop_with_include(self):
        self.populate_file('main.yaml', '''---
        include: 
            file: print_file.yaml
            as: print
        variables:
            list: [{'file': 'a', 'value': 1}, {'file': 'b', 'value': 2}, {'file': 'c', 'value': 3}]
        steps:
            - loop: 
                foreach:
                    in: '{{ list }}'
                    do:
                        run: 
                            include: 'print'
                            variables: 
                                data: '{{ ITEM["value"] }}'
                                filename: '{{ ITEM["file"] }}'
        ''')
        self.populate_file('print_file.yaml', '''---
         steps:
            - echo: {from: '{{ data }}', to: '{{ filename }}'}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'a'), '1'))
        self.assertTrue(check_file(join(self.test_dir, 'b'), '2'))
        self.assertTrue(check_file(join(self.test_dir, 'c'), '3'))



import os
from os.path import join

import requests_mock

from catcher.core.runner import Runner
from catcher.utils.file_utils import read_file
from test.abs_test_class import TestClass
from test.test_utils import check_file


class RunTest(TestClass):
    def __init__(self, method_name):
        super().__init__('run_test', method_name)

    # test run one action
    def test_run_one_action(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: 'test', to: main.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'main.output'), 'test'))

    # run action and register variable in it
    def test_run_register(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: 'user-{{ RANDOM_STR }}', register: {uuid: '{{ OUTPUT }}'}}
            - echo: {from: '{{ uuid }}', to: main.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        file = read_file(join(self.test_dir, 'main.output'))
        self.assertTrue(file.startswith('user'))
        self.assertTrue('{{' not in file)  # template was filled in

    # test run several actions
    def test_run_several_actions(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: 'test1', to: main1.output}
            - echo: {from: 'test2', to: main2.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'main1.output'), 'test1'))
        self.assertTrue(check_file(join(self.test_dir, 'main2.output'), 'test2'))

    # test run ignore errors
    @requests_mock.mock()
    def test_run_ignore_errors(self, m):
        m.get('http://test.com', status_code=500)
        self.populate_file('main.yaml', '''---
        steps:
            - http: {get: {url: 'http://test.com', response_code: 200}, ignore_errors: true}
            - echo: {from: 'test2', to: main2.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'main2.output'), 'test2'))

    def test_run_named(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: 
                from: 'user-{{ RANDOM_STR }}'
                register: {uuid: '{{ OUTPUT }}'}
                name: 'first'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_run_named_variables(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: 'user-{{ RANDOM_STR }}', register: {uuid: '{{ OUTPUT }}'}}
            - echo: 
                from: '{{ uuid }}'
                to: main.output
                name: 'Save user {{ uuid }} to main.output'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_run_system_variables(self):
        os.environ["SECRET_PASSWORD"] = "123"
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: '{{ SECRET_PASSWORD }}', to: sys_env.output}
        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        None,
                        system_environment=dict(os.environ))
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'sys_env.output'), '123'))

    def test_run_skip_if_short(self):
        self.populate_file('main.yaml', '''---
                variables:
                    no_output: true
                steps:
                    - echo:
                        from: 'hello world'
                        to: main1.output
                        skip_if: '{{ no_output }}'
                    - echo: {from: 'test2', to: main2.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(not os.path.exists(join(self.test_dir, 'main1.output')))
        self.assertTrue(check_file(join(self.test_dir, 'main2.output'), 'test2'))

    def test_run_skip_if_long(self):
        self.populate_file('main.yaml', '''---
                        variables:
                            user_name: 'test'
                        steps:
                            - echo:
                                from: 'hello world'
                                to: main1.output
                                skip_if:
                                    equals: {the: '{{ user_name }}', is: 'test'}
                            - echo: {from: 'test2', to: main2.output}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(not os.path.exists(join(self.test_dir, 'main1.output')))

    def test_run_skip_if_multiple(self):
        self.populate_file('main.yaml', '''---
                                variables:
                                    list: ['a', 'b', 'c']
                                steps:
                                    - echo:
                                        from: 'hello world'
                                        to: main1.output
                                        skip_if:
                                            or:
                                                - contains: {the: '1', in: '{{ list }}'}
                                                - equals: {the: '{{ list[0] }}', is: 'a'}
                                                - contains: {the: 'b', in: '{{ list }}'}
                                    - echo: {from: 'test2', to: main2.output}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(not os.path.exists(join(self.test_dir, 'main1.output')))

    def test_run_ignore_test(self):
        self.populate_file('main.yaml', '''---
                ignore: true # for some reason this test is not working
                steps:
                    - check: {equals: {the: true, is: false}}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_ignore_condition(self):
        os.environ['CLOUD'] = 'AWS'
        self.populate_file('main.yaml', '''---
                        ignore: 
                            equals: {the: '{{ CLOUD }}', is: 'AWS'}
                        steps:
                            - check: {equals: {the: true, is: false}}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, system_environment=dict(os.environ))
        self.assertTrue(runner.run_tests())

        os.environ['CLOUD'] = 'Marathon'
        self.populate_file('main.yaml', '''---
                                ignore: 
                                    equals: {the: '{{ CLOUD }}', is: 'AWS'}
                                steps:
                                    - check: {equals: {the: true, is: false}}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, system_environment=dict(os.environ))
        self.assertFalse(runner.run_tests())




import os
from os.path import join

from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from test.test_utils import check_file


class VariablesTest(TestClass):
    def __init__(self, method_name):
        super().__init__('variables_test', method_name)

    # newly registered variables override everything
    def test_register_override_cmd(self):
        self.populate_file('main.yaml', '''---
        variables:
            foo: baz
        steps:
            - echo: 
                actions: 
                    - {from: 'bar', register: {foo: '{{ OUTPUT }}'}}
                    - {from: '{{ foo }}', to: one.output}
            
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, cmd_env={'foo': 'bad'})
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'bar'))

    # variables, passed via cmd `e` options should override all variables
    def test_cmd_override_all(self):
        self.populate_file('inventory.yml', '''---
                foo2: baz
                ''')
        self.populate_file('main.yaml', '''---
        variables:
            foo: baz
        steps:
            - echo: 
                actions: 
                    - {from: '{{ foo }}', to: one.output}
                    - {from: '{{ foo2 }}', to: two.output}

        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        join(self.test_dir, 'inventory.yml'),
                        cmd_env={'foo': 'bad', 'foo2': 'bad'})
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'bad'))
        self.assertTrue(check_file(join(self.test_dir, 'two.output'), 'bad'))

    # test computed variables available later
    def test_computed_available_later(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: 
                actions: 
                    - {from: 'hello', register: {foo: 'bar'}}
                    - {from: '{{ foo }}', to: one.output}
            
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'bar'))

    # test computed available in the other step
    def test_computed_available_other_step(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: 'hello', register: {foo: 'bar'}}
            - echo: {from: '{{ foo }}', to: one.output}
            
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), 'bar'))

    # test var can be computed from the template
    def test_computed_with_template(self):
        self.populate_file('main.yaml', '''---
        variables:
            path: /home/user
            test: main.yml
        steps:
            - echo: {from: 'hello', register: {foo: '{{ path }}/{{ test }}'}}
            - echo: {from: '{{ foo }}', to: one.output}
            
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'one.output'), '/home/user/main.yml'))

    # test if env vars included from inventory
    def test_env_vars_included(self):
        os.environ['FOO'] = '1'
        self.populate_file('main.yaml', '''---
        steps:
            - check: {equals: {the: '{{ FOO }}', is: '1'}}

        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        None,
                        system_environment=dict(os.environ))
        self.assertTrue(runner.run_tests())

    # env var should be overridden with any other var
    def test_var_override_env(self):
        os.environ['FOO'] = '1'
        self.populate_file('main.yaml', '''---
        variables:
            FOO: 2
        steps:
            - check: {equals: {the: '{{ FOO }}', is: '2'}}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_env_var_in_vars(self):
        os.environ['FOO'] = '1'
        self.populate_file('main.yaml', '''---
                variables:
                    foo: '{{ FOO }}'
                steps:
                    - check: {equals: {the: '{{ foo }}', is: '1'}}

                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, system_environment=dict(os.environ))
        self.assertTrue(runner.run_tests())

    def test_var_in_run_as_obj(self):
        self.populate_resource('test.csv', "email,id\n"
                                           "{%- for user in users %}\n"
                                           "{{ user.email }},{{ user.id }}\n"
                                           "{%- endfor -%}"
                               )
        self.populate_file('main.yaml', '''---
                                variables:
                                    users:
                                        - email: 'first@test.de'
                                          id: 1
                                        - email: 'second@test.de'
                                          id: 2
                                include: 
                                    file: one.yaml
                                    as: one
                                steps:
                                    - run: 
                                        include: 'one'
                                        variables:
                                            users: '{{ users[:1] }}'
                                ''')
        self.populate_file('one.yaml', '''---
                                steps:
                                    - echo: {from_file: 'test.csv', to: res.output}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'res.output'), 'email,id\nfirst@test.de,1'))

    def test_json_in_variable(self):
        self.populate_file('inventory.yml', '''---
        complex_conf:
          field: 'value'
          json_field: '{"host": "http://minio:9000","aws_access_key_id":"minio","aws_secret_access_key":"minio123"}'
        ''')
        self.populate_file('main.yaml', '''---
                steps:
                    - echo: {from: '{{ complex_conf }}'}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), join(self.test_dir, 'inventory.yml'))
        self.assertTrue(runner.run_tests())



from os.path import join

import pytest
import requests_mock

from catcher.core.runner import Runner
from test.abs_test_class import TestClass


class HttpTest(TestClass):
    def __init__(self, method_name):
        super().__init__('http', method_name)

    @requests_mock.mock()
    def test_status_code_get(self, m):
        m.get('http://test.com', status_code=500)
        self.populate_file('main.yaml', '''---
            steps:
                - http: {get: {url: 'http://test.com', response_code: 200}}
            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertFalse(runner.run_tests())

    @requests_mock.mock()
    def test_status_code_ok(self, m):
        m.get('http://test.com', status_code=200)
        self.populate_file('main.yaml', '''---
            steps:
                - http: {get: {url: 'http://test.com', response_code: 200}}
            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @requests_mock.mock()
    def test_get_response(self, m):
        m.get('http://test.com', json={'uuid': 'qwerty', 'checked': True})
        self.populate_file('main.yaml', '''---
            steps:
                - http: 
                    get: 
                        url: 'http://test.com'
                        response_code: 200
                    register: {checked: '{{ OUTPUT.checked }}'}
                - check: '{{ checked }}'
            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @requests_mock.mock()
    def test_post_response(self, m):
        m.post('http://test.com', text='Error message: no such id')
        self.populate_file('main.yaml', '''---
            steps:
                - http: 
                    post: 
                        url: 'http://test.com'
                        body: {'id': 1234, 'action': 'fee'}
                    register: {reply: '{{ OUTPUT }}'}
                - check:
                    contains: {'the': 'Error', in: '{{ reply }}'}
            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @requests_mock.mock()
    def test_post_with_vars(self, m):
        m.post('http://test.com', json={'id': 1234, 'action': 'fee'})
        self.populate_file('main.yaml', '''---
            variables:
                id: 1234
            steps:
                - http: 
                    post: 
                        url: 'http://test.com'
                        body: {'id': '{{ id }}', 'action': 'fee'}
                    register: {reply: '{{ OUTPUT.id }}'}
                - check:
                    equals: {the: '{{ reply }}', is: '{{ id }}'}
            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @requests_mock.mock()
    def test_post_from_file(self, m):
        m.post('http://test.com', text="loaded ok")
        self.populate_file('answers.json', '''
        {
            "answers":[
                {"id": 1,"response": true},
                {"id": 2,"response": false},
                {"id": 3,"response": true},
                {"id": 4,"response": false},
                {"id": 5,"response": true}
            ]
        }
        ''')

        self.populate_file('main.yaml', '''---
            steps:
                - http: 
                    post: 
                        url: 'http://test.com'
                        body_from_file: "'''
                           + join(self.test_dir, 'answers.json') + '''"
                    register: {reply: '{{ OUTPUT }}'}
                - check:
                    equals: {the: '{{ reply }}', is: 'loaded ok'}
            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_should_fail(self):
        self.populate_file('main.yaml', '''---
                    steps:
                        - http: 
                            get: 
                                url: 'http://undefined'
                                should_fail: true
                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @requests_mock.mock()
    def test_await_2xx(self, m):
        m.get('http://test.com', status_code=201)
        self.populate_file('main.yaml', '''---
                    steps:
                        - http: {get: {url: 'http://test.com', response_code: 2xx}}
                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @requests_mock.mock()
    def test_await_range(self, m):
        m.get('http://test.com', status_code=201)
        self.populate_file('main.yaml', '''---
                            steps:
                                - http: {get: {url: 'http://test.com', response_code: 200-300}}
                            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @requests_mock.mock()
    def test_await_range_xx(self, m):
        m.get('http://test.com', status_code=201)
        self.populate_file('main.yaml', '''---
                                    steps:
                                        - http: {get: {url: 'http://test.com', response_code: 2xx-3xx}}
                                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @requests_mock.mock()
    def test_send_json_via_headers(self, m):
        adapter = m.post('http://test.com')

        self.populate_file('main.yaml', '''---
                                    variables:
                                        body: {'foo': 'bar'}
                                    steps:
                                        - http: 
                                            post: 
                                                url: 'http://test.com'
                                                headers: {Content-Type: 'application/json'}
                                                body: '{{ body }}'
                                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertEqual({'foo': 'bar'}, adapter.last_request.json())

    @requests_mock.mock()
    def test_send_json_directly(self, m):
        adapter = m.post('http://test.com')

        self.populate_file('main.yaml', '''---
                                    variables:
                                        body: {'foo': 'bar'}
                                    steps:
                                        - http: 
                                            post: 
                                                url: 'http://test.com'
                                                body: '{{ body |tojson }}'
                                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertEqual({'foo': 'bar'}, adapter.last_request.json())

    @requests_mock.mock()
    def test_upload_file(self, m):
        self.populate_file('main.yaml', '''---
                                    steps:
                                        - http: 
                                            post: 
                                                url: 'http://test.com'
                                                files:
                                                    file: 'foo.csv'
                                                    type: 'text/csv'
                                    ''')
        self.populate_resource('foo.csv', "one,two\n"
                                          "three,four"
                               )

        adapter = m.post('http://test.com')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue('Content-Type: text/csv' in adapter.last_request.text)
        self.assertTrue("one,two\n"
                        "three,four" in adapter.last_request.text)

    @requests_mock.mock()
    def test_upload_multiple(self, m):
        self.populate_file('main.yaml', '''---
                                    steps:
                                        - http: 
                                            post: 
                                                url: 'http://test.com'
                                                files:
                                                    - file1: 'one.json'
                                                      type: 'application/json'
                                                    - file2: 'foo.csv'
                                                      type: 'text/csv'
                                    ''')
        self.populate_resource('one.json', "{\"key\":\"value\"}")
        self.populate_resource('foo.csv', "one,two\n"
                                          "three,four"
                               )
        adapter = m.post('http://test.com')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue('Content-Type: text/csv' in adapter.last_request.text)
        self.assertTrue('Content-Type: application/json' in adapter.last_request.text)
        self.assertTrue("{\"key\":\"value\"}" in adapter.last_request.text)
        self.assertTrue("one,two\n"
                        "three,four" in adapter.last_request.text)

    @requests_mock.mock()
    def test_upload_template(self, m):
        self.populate_file('main.yaml', '''---
                                    variables:
                                        var: 'one'
                                    steps:
                                        - http: 
                                            post: 
                                                url: 'http://test.com'
                                                files:
                                                    file: 'foo.csv'
                                                    type: 'text/csv'
                                    ''')
        self.populate_resource('foo.csv', "{{var}},two\n"
                                          "three,four"
                               )

        adapter = m.post('http://test.com')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue('Content-Type: text/csv' in adapter.last_request.text)
        self.assertTrue("one,two\n"
                        "three,four" in adapter.last_request.text)

    @pytest.mark.skip(reason="Uses external service. Shouldn't be run automatically")
    def test_use_cookies(self):
        self.populate_file('main.yaml', '''---
                            steps:
                                - http: 
                                    get: 
                                        url: 'http://httpbin.org/cookies/set/sessioncookie/123456789'
                                - http:
                                    get:
                                        url: 'http://httpbin.org/cookies'
                                    register: {reply: '{{ OUTPUT }}'}
                                - check:
                                    equals: {the: '{{ reply.cookies.sessioncookie }}', is: '123456789'}
                            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    @pytest.mark.skip(reason="Uses external service. Shouldn't be run automatically")
    def test_change_sessions(self):
        self.populate_file('main.yaml', '''---
                                    steps:
                                        - http: 
                                            get: 
                                                url: 'http://httpbin.org/cookies/set/sessioncookie/123456789'
                                                session: 'one'
                                        - http: 
                                            get: 
                                                url: 'http://httpbin.org/cookies/set/sessioncookie/987654321'
                                                session: 'two'
                                        - http:
                                            get:
                                                url: 'http://httpbin.org/cookies'
                                                session: 'one'
                                            register: {reply: '{{ OUTPUT }}'}
                                        - check:
                                            equals: {the: '{{ reply.cookies.sessioncookie }}', is: '123456789'}
                                        - http:
                                            get:
                                                url: 'http://httpbin.org/cookies'
                                                session: 'two'
                                            register: {reply: '{{ OUTPUT }}'}
                                        - check:
                                            equals: {the: '{{ reply.cookies.sessioncookie }}', is: '987654321'}
                                        - http: 
                                            get: 
                                                url: 'http://httpbin.org/cookies/set/sessioncookie/123456789'
                                                session: null
                                        - http:
                                            get:
                                                url: 'http://httpbin.org/cookies'
                                                session: null
                                            register: {reply: '{{ OUTPUT }}'}
                                        - check:   
                                            equals: {the: '{{ reply.cookies }}', is: '{}'}
                                        
                                    ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())



"""Catcher - Microservices automated test tool.

Usage:
  catcher [-i INVENTORY] <tests> [-l LEVEL] [-e VARS...] [-m MODS...]... [-r RES] [-p FORMAT] [-s SYS_ENV] [-f FILTER]... [--q | --qq] [--no-color]
  catcher -v | --version
  catcher -h | --help

Options:
  -h --help                          show this help message and exit
  -v --version                       print version and exit
  -l LEVEL --log-level LEVEL         set log level. Options: debug, info, warning, error, critical [default: info]
  -i INVENTORY --inventory INVENTORY inventory file with environment configuration
  -e VARIABLE --environment VARIABLE set variable (will override inventory).
  -s SYS_ENV --system_env SYS-ENV    use system environment variables as variables [default: true]
  -m MODULES --modules MODULES       specify directories or python packages to search for external modules
  -r RESOURCES --resources RESOURCES set the resources dir [default: ./resources]
  -p FORMAT --format FORMAT          set the format (json) for the resulting file, which includes all steps
                                     execution results, variables and outputs. It is created in the `reports` directory.
                                     Is not created by default.
  -f FILTER --filter FILTER          Path to python file with custom filters implementation or python module's path if
                                     installed in the system.
  --q                                Do not print steps output
  --qq                               Do not print steps and tests output
  --no-color                         Do not use colorful output.
"""
import os
import sys
from distutils.util import strtobool

from docopt import docopt, DocoptExit

from catcher import APPVSN
from catcher.core.runner import Runner
from catcher.utils import logger
from catcher.utils.logger import warning
from catcher.utils.module_utils import load_external_actions
from colorama import init

init()


def main(args=None):
    try:
        arguments = docopt(__doc__, argv=args, version=APPVSN)
    except DocoptExit as usage:
        print(usage)
        sys.exit(1)
    path = os.getcwd()
    logger.configure(arguments['--log-level'], not arguments['--no-color'])
    result = run_tests(path, arguments)
    if result:
        sys.exit(0)
    else:
        sys.exit(1)


def run_tests(path: str, arguments: dict):
    file_or_dir = arguments['<tests>']
    inventory = arguments['--inventory']
    environment = arguments['--environment']
    modules = arguments['--modules']
    resources = arguments['--resources']
    output_format = arguments['--format']
    filters = arguments['--filter']
    use_sys_vars = arguments['--system_env']
    output = 'full' if not arguments['--q'] else 'limited'
    if arguments['--qq']:
        output = 'final'
    if strtobool(use_sys_vars):
        sys_vars = dict(os.environ)
    else:
        sys_vars = None
    __load_modules(modules)
    runner = Runner(path, file_or_dir, inventory,
                    modules=modules,
                    cmd_env=__env_to_variables(environment),
                    resources=resources,
                    system_environment=sys_vars,
                    output_format=output_format,
                    filter_list=filters)
    return runner.run_tests(output=output)


def __env_to_variables(environment: list) -> dict:
    variables = {}
    for env in environment:
        if '=' not in env:
            warning('Skip not kv env param ' + env)
        else:
            [k, v] = env.split('=')
            variables[k] = v
    return variables


def __load_modules(modules):
    """
    Load catcher core modules, carcher-modules extensions (if available) and try to load all modules from cmd args
    """
    load_external_actions('catcher.steps')
    load_external_actions('catcher_modules')
    [load_external_actions(m) for m in modules]


if __name__ == "__main__":
    main()



from os.path import join

from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from test.test_utils import check_file


class ChecksTest(TestClass):
    def __init__(self, method_name):
        super().__init__('checks_test', method_name)

    def setUp(self):
        super().setUp()

    # test read file with echo
    def test_read_file(self):
        self.populate_resource('debug.input', '123')
        self.populate_file('main.yaml', '''---
        variables:
            foo: true
        steps:
            - echo: {from_file: debug.input, register: {user_email: '{{ OUTPUT }}'}}
            - check: {equals: {the: '{{ user_email }}', is: '123'}}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test echo to file
    def test_write_file(self):
        self.populate_file('main.yaml', '''---
                variables:
                    user_email: 123
                steps:
                    - echo: {from: '{{ user_email }}', to: 'debug.output'}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'debug.output'), '123'))

    # test write to a variable
    def test_write_variable(self):
        self.populate_file('main.yaml', '''---
                        variables:
                            user_email: 123
                        steps:
                            - echo: {from: '{{ user_email }}', register: {'foo': '{{ OUTPUT }}'} }
                            - check: {equals: {the: '{{ foo }}', is: '123'}}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_read_write_to_file_template(self):
        self.populate_resource('foo/baz/bar.json', '123')
        self.populate_file('main.yaml', '''---
                        variables:
                            filename: foo/baz/bar.json
                        steps:
                            - echo: {from_file: '{{ filename }}', to: 'resources/{{ filename.replace("/baz/", "/bar/")}}'}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'resources/foo/bar/bar.json'), '123'))

    def test_read_python_file(self):
        self.populate_resource('config.py', '''
my_var = 123
my_dict = dict(first=1, second=2)
my_list = [1,2,3]        
        ''')
        self.populate_file('main.yaml', '''---
                                steps:
                                    - echo:
                                        from_file: "config.py"
                                        register: {"data": "{{ OUTPUT }}"}
                                    - check: {equals: {the: '{{ data.my_var }}', is: 123}}
                                    - check: {equals: {the: '{{ data.my_dict.first }}', is: 1}}
                                    - check: {equals: {the: '{{ data.my_dict.second }}', is: 2}}
                                    - check: {equals: {the: '{{ data.my_list[0] }}', is: 1}}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())


import json
from copy import deepcopy
from datetime import datetime
from catcher.modules.formatter import formatter_factory
from catcher.utils import file_utils


class LogStorage:
    def __init__(self, output_format) -> None:
        self._data = []
        self._current_test = None
        self._format = output_format
        self.nesting_counter = 0

    def test_start(self, test: str, test_type='test'):
        self._current_test = {'start_time': datetime.now().strftime("%Y-%m-%d %H:%M:%S"), 'file': test,
                              'type': test_type, 'output': [], 'status': 'running'}

    def test_end(self, test, success: bool, output: str = None, test_type='test', end_comment=None):
        if self._current_test:
            if success:
                self._current_test['status'] = 'OK'
            else:
                if output:
                    self._current_test['status'] = output
                else:
                    self._current_test['status'] = 'FAIL'
        else:
            self._current_test = {}  # to avoid NPE on the next line
        self._current_test['comment'] = end_comment
        self._data += [{**{'end_time': datetime.now().strftime("%Y-%m-%d %H:%M:%S")}, **self._current_test}]
        self._current_test = None

    def test_parse_fail(self, test: str, output: str):
        self._data += [{'end_time': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        'start_time': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                        'file': test,
                        'comment': None,
                        'type': 'test',
                        'output': output,
                        'status': 'FAIL'}]

    def nested_test_in(self):
        """
        If test is run from include
        """
        self.nesting_counter += 1

    def nested_test_out(self):
        """
        If test is run from include
        """
        self.nesting_counter -= 1

    def new_step(self, step, variables):
        self._current_test['output'] += [{'time': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                                          'step': self.clean_step_def(step),
                                          'variables': self.clean_vars(variables),
                                          'nested': self.nesting_counter}]

    def step_end(self, step, variables, output: str = None, success: bool = True):
        self._current_test['output'] += [{'time': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                                          'step': self.clean_step_def(step),
                                          'variables': self.clean_vars(variables),
                                          'nested': self.nesting_counter,
                                          'success': success,
                                          'output': output}]

    def output(self, level, output):
        if self._current_test:
            self._current_test['output'] += [{'time': datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                                              'data': output,
                                              'level': level}]
        else:
            self._data += [{'time': datetime.now().strftime("%Y-%m-%d %H:%M:%S"), 'data': output, 'level': level}]

    def write_report(self, path):
        file_utils.ensure_dir(path)
        formatter = formatter_factory(self._format)
        formatter.format(path, self._data)

    def print_summary(self, path):
        from catcher.utils import logger
        tests = [t for t in self._data if t.get('type') == 'test']
        out_string = self.calculate_statistics(tests)
        for test in tests:
            out_string += '\nTest {}: '.format(file_utils.cut_path(path, test['file']))
            # select only step's ends, which belongs to the current test (excluding registered includes run)
            step_finish_only = [o for o in test['output'] if 'success' in o and o['nested'] == 0]
            if test['status'] == 'OK' and test['comment'] != 'Skipped':
                out_string += logger.green('pass')
            elif test['comment'] == 'Skipped':
                out_string += logger.yellow('skipped')
            else:
                out_string += logger.red('fail') + ', on step {}'.format(len(step_finish_only))
        logger.info(out_string)

    @staticmethod
    def calculate_statistics(tests):
        from catcher.utils import logger
        ok = len([t for t in tests if t['status'] == 'OK' and t['comment'] != 'Skipped'])
        skipped = len([t for t in tests if t['comment'] == 'Skipped'])
        fail = len(tests) - (ok + skipped)
        percent = (ok + skipped) / len(tests) * 100 if len(tests) > 0 else 0
        out_string = 'Test run {}.'.format(logger.blue(str(len(tests))))
        out_string += ' Success: ' + logger.green(str(ok)) + ', Fail: '
        if fail > 0:
            out_string += logger.red(str(fail))
        else:
            out_string += logger.green(str(fail))
        if skipped:
            out_string += ', Skipped: ' + logger.yellow(str(skipped))
        out_string += '. Total: '
        fun = logger.red
        if percent >= 100:
            fun = logger.green
        elif percent >= 66:
            fun = logger.light_green
        elif percent >= 33:
            fun = logger.yellow
        out_string += fun('{:.0f}%'.format(percent))
        return out_string

    def clean_step_def(self, data: dict):
        """ Keep only step definition """
        [step_name] = data.keys()
        step = data[step_name]
        if isinstance(step, str):  # simple step like {'echo': 'hello'}
            return data
        step_def = self.clean_not_renderable_recursive(step)
        return {step_name: step_def}

    def clean_not_renderable_recursive(self, data):
        if not hasattr(data, 'copy'):  # list with primitives [1, 2, 3] should be ignored
            return data
        step_def = data.copy()
        for k, v in data.items():
            if isinstance(v, dict):
                step_def[k] = self.clean_not_renderable_recursive(v)
            elif isinstance(v, list):
                step_def[k] = [self.clean_not_renderable_recursive(t) for t in v]
            elif not self.is_jsonable(v):
                del step_def[k]
        return step_def

    @staticmethod
    def clean_vars(variables: dict):
        """ Clean _get_action(s) non-json serializable functions """
        return deepcopy({k: v for k, v in variables.items() if not callable(v)})

    @staticmethod
    def is_jsonable(x):
        try:
            json.dumps(x)
            return True
        except (TypeError, OverflowError):
            return False


class EmptyLogStorage(LogStorage):
    """
    The default implementation. Does nothing.
    """

    def new_step(self, step, variables):  # ignore variables to free memory
        super().new_step(step, {})

    def step_end(self, step, variables, output: str = None, success: bool = True):  # ignore variables to free memory
        super().step_end(step, {}, output, success)

    def output(self, level, output):
        pass

    def write_report(self, path):
        pass



from catcher.utils import module_utils
from catcher.utils.singleton import Singleton
from catcher.utils.logger import info, debug


class FiltersFactory(metaclass=Singleton):
    def __init__(self, custom_modules=None) -> None:
        super().__init__()
        self._filters = {}
        self._functions = {}
        if not custom_modules:
            custom_modules = []
        custom_modules.append('catcher.modules.filter_impl.bifs')
        self._import_custom(custom_modules)

    @property
    def filters(self):
        return self._filters

    @property
    def functions(self):
        return self._functions

    def _import_custom(self, custom_modules):
        """
        Will import all functions started with name 'function_' to global functions and all functions started with
        name 'filter_' to filters. They will be available in templates.
        'function_' and 'filter_' prefixes will be removed from function names during the import.
        :param custom_modules: list of paths to python files where filters of functions are defined.
        """
        for filter_module in custom_modules:
            info('Loading {}'.format(filter_module))
            funs = module_utils.get_all_functions(filter_module)
            for fun_name, fun in funs.items():
                if fun_name.startswith('function'):
                    import_name = '_'.join(fun_name.split('_')[1:])
                    debug('Adding function {}'.format(import_name))
                    self._functions[import_name] = fun
                elif fun_name.startswith('filter'):
                    import_name = '_'.join(fun_name.split('_')[1:])
                    debug('Adding filter {}'.format(import_name))
                    self._filters[import_name] = fun



from os.path import join

import pytest
from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from test.test_utils import read_file, check_file


class ShTest(TestClass):
    def __init__(self, method_name):
        super().__init__('sh_test', method_name)

    def test_list_directory(self):
        self.populate_file('main.yaml', '''---
        steps:
            - sh:
                command: 'ls -la'
                path: '{{ CURRENT_DIR }}'
                register: {content: '{{ OUTPUT }}'}
            - echo: {from: '{{ content }}', to: 'debug.output'}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        file = read_file(join(self.test_dir, 'debug.output'))
        self.assertTrue('resources' in file)

    # try to delete non existent directory and fail test with non zero return code.
    def test_run_negative(self):
        self.populate_file('main.yaml', '''---
                steps:
                    - sh:
                        command: 'rm foobarbaz'
                    - echo: {from: '{{ content }}', to: 'debug.output'}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertFalse(runner.run_tests())

    @pytest.mark.skip(reason="Disabled for CI, as it may run tests in docker")
    def test_run_in_docker(self):
        self.populate_file('main.yaml', '''---
                        variables:
                            docker: true
                        steps:
                            - sh:
                                command: "grep 'docker|lxc' /proc/1/cgroup"
                                return_code: 1
                                ignore_errors: true
                                register: {docker: false}
                            - echo: {from: '{{ docker }}', to: 'debug.output'}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())
        self.assertTrue(check_file(join(self.test_dir, 'debug.output'), 'False'))



from os.path import join

from catcher.core.runner import Runner
from test.abs_test_class import TestClass


class ChecksTest(TestClass):
    def __init__(self, method_name):
        super().__init__('checks_test', method_name)

    # test simple equals
    def test_check_short_form(self):
        self.populate_file('main.yaml', '''---
        variables:
            foo: true
        steps:
            - check: '{{ foo }}'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test computed var equals constant value
    def test_computed_equals_constant(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: 'hello', register: {'foo': 'value'}}
            - check: {equals: {the: '{{ foo }}', is: 'value'}}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test computed var equals other var
    def test_computed_equals_variable(self):
        self.populate_file('main.yaml', '''---
        variables:
            foo: bar
            foo2: bar
        steps:
            - echo: {from: '{{ foo }}'}
            - check: {equals: {the: 'bar', is: '{{ foo }}'}}
            - check: {equals: {the: '{{ foo }}', is: 'bar'}}
            - check: {equals: {the: '{{ foo2 }}', is: '{{ foo }}'}}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test if any of the list's elements is equal to var
    def test_any_equal(self):
        self.populate_file('main.yaml', '''---
        variables:
            foo: [bar, baz]
        steps:
            - echo: {from: '{{ foo }}'}
            - check: 
                any:
                    of: '{{ foo }}'
                    equals: 'baz'
            - check: 
                any:
                    of: [bar, baz]
                    equals: 'baz'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test if all of the list's elements is equal to var
    def test_all_equal(self):
        self.populate_file('main.yaml', '''---
        variables:
            foo: [1, 1]
        steps:
            - echo: {from: '{{ foo }}'}
            - check: 
                all:
                    of: '{{ foo }}'
                    equals: 1
            - check: 
                all: 
                    of: [1,1]
                    equals: 1
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test if `contains` work with list/dict
    def test_contains(self):
        self.populate_file('main.yaml', '''---
        variables:
            list: ['a', 'b', 'c']
            dict: {'a': 1, 'b': 2, 'c': 3}
        steps:
            - check: 
                contains: {the: 'a', in: '{{ list }}'}
            - check: 
                contains: {the: 'a', in: '{{ dict }}'}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    def test_contains_string(self):
        self.populate_file('main.yaml', '''---
                variables:
                    word: 'hello'
                    word2: 'incorrect'
                    phrase: 'first hello world'
                steps:
                    - check: 
                        contains: {the: '{{ word }}', in: '{{ phrase }}'}
                    - check: 
                        contains: {the: '{{ word2 }}', not_in: '{{ phrase }}'}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test if `or` works
    def test_or(self):
        self.populate_file('main.yaml', '''---
        variables:
            list: ['a', 'b', 'c']
        steps:
            - check: 
                or: 
                    - contains: {the: '1', in: '{{ list }}'}
                    - equals: {the: '{{ list[0] }}', is: 'a'}
                    - contains: {the: 'b', in: '{{ list }}'}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test if `and` works
    def test_and(self):
        self.populate_file('main.yaml', '''---
        variables:
            list: ['a', 'b', 3]
        steps:
            - check: 
                and: 
                    - contains: {the: 'a', in: '{{ list }}'}
                    - equals: {the: '{{ list[1] }}', is: 'b'}
                    - equals: {the: '{{ list[2] > 2 }}', is: true}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test if inverted operators work
    def test_negative(self):
        self.populate_file('main.yaml', '''---
        variables:
            list: ['d', 'c', 1]
        steps:
            - check: 
                and: 
                    - contains: {the: 'a', not_in: '{{ list }}'}
                    - equals: {the: '{{ list[1] }}', is_not: 'b'}
                    - equals: {the: '{{ list[2] > 2 }}', is_not: true}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test if ITEM is available in list
    def test_items_iteration(self):
        self.populate_file('main.yaml', '''---
        variables:
            list: [{n: 1, k: 'a'}, {n: 2, k: 'a'}, {n: 3, k: 'a'}]
        steps:
            - check: 
                all:
                    of: '{{ list }}'
                    equals: {the: '{{ ITEM.k }}', is: 'a'}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())

    # test if dates with dashes are not calculated as math (old dates, i.e. 1973-03-10 are still calculated as math)
    def test_compate_dates(self):
        self.populate_file('main.yaml', '''---
        steps:
            - check:
                equals: {the: '2020-03-11', is_not: '2020-11-03' }
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        self.assertTrue(runner.run_tests())



import json
from os import listdir
from os.path import join, isfile

from catcher.core.runner import Runner
from test.abs_test_class import TestClass


class OutputTest(TestClass):
    def __init__(self, method_name):
        super().__init__('output_test', method_name)

    def test_run_multiple_steps(self):
        self.populate_file('main.yaml', '''---
        variables:
            foo: [bar, baz]
        steps:
            - echo: {from: '{{ foo }}'}
            - check: 
                any:
                    of: '{{ foo }}'
                    equals: 'baz'
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, output_format='json')
        self.assertTrue(runner.run_tests())
        reports = [f for f in listdir(join(self.test_dir, 'reports'))
                   if isfile(join(self.test_dir, 'reports', f)) and f.startswith('report')]
        self.assertEqual(1, len(reports))
        with open(join(self.test_dir, 'reports', reports[0]), 'r') as fp:
            obj = json.load(fp)
            self.assertEqual(1, len(obj))
            report = obj[0]
            self.assertEqual('test', report['type'])
            self.assertEqual(join(self.test_dir, 'main.yaml'), report['file'])
            output = report['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual(4, len(steps))  # 4 step entry for 2 steps (before & after)
            self.assertEqual('echo', list(steps[0]['step'].keys())[0])
            self.assertTrue(steps[1]['success'])
            self.assertEqual('check', list(steps[2]['step'].keys())[0])
            self.assertTrue(steps[3]['success'])
            self.assertEqual('OK', report['status'])

    def test_step_failed(self):
        self.populate_file('main.yaml', '''---
                variables:
                    foo: [bar]
                steps:
                    - echo: {from: '{{ foo }}'}
                    - check: 
                        any:
                            of: '{{ foo }}'
                            equals: '404'
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, output_format='json')
        self.assertFalse(runner.run_tests())
        reports = [f for f in listdir(join(self.test_dir, 'reports'))
                   if isfile(join(self.test_dir, 'reports', f)) and f.startswith('report')]
        self.assertEqual(1, len(reports))
        with open(join(self.test_dir, 'reports', reports[0]), 'r') as fp:
            obj = json.load(fp)
            self.assertEqual(1, len(obj))
            report = obj[0]
            self.assertEqual('test', report['type'])
            self.assertEqual(join(self.test_dir, 'main.yaml'), report['file'])
            output = report['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual(4, len(steps))  # 4 step entry for 2 steps (before & after)
            self.assertEqual('echo', list(steps[0]['step'].keys())[0])
            self.assertTrue(steps[1]['success'])
            self.assertEqual('check', list(steps[2]['step'].keys())[0])
            self.assertFalse(steps[3]['success'])
            self.assertTrue(self._compare_operations(steps[3]['output'],
                                                     "operation {'any': {'of': '['bar']', 'equals': '404'}} failed",
                                                     "operation {'any': {'equals': '404', 'of': '['bar']'}} failed"))
            self.assertTrue(self._compare_operations(report['status'],
                                                     "operation {'any': {'of': '['bar']', 'equals': '404'}} failed",
                                                     "operation {'any': {'equals': '404', 'of': '['bar']'}} failed"))

    def test_register_variable(self):
        self.populate_file('main.yaml', '''---
                        steps:
                            - echo: {from: 'val', register: {user: '{{ OUTPUT }}'}}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, output_format='json')
        self.assertTrue(runner.run_tests())
        reports = [f for f in listdir(join(self.test_dir, 'reports'))
                   if isfile(join(self.test_dir, 'reports', f)) and f.startswith('report')]
        self.assertEqual(1, len(reports))
        with open(join(self.test_dir, 'reports', reports[0]), 'r') as fp:
            obj = json.load(fp)
            self.assertEqual(1, len(obj))
            report = obj[0]
            self.assertEqual('test', report['type'])
            self.assertEqual(join(self.test_dir, 'main.yaml'), report['file'])
            output = report['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual(2, len(steps))  # 2 step entry for 1 step (before & after)
            self.assertEqual('echo', list(steps[0]['step'].keys())[0])
            self.assertEqual(sorted(['CURRENT_DIR', 'RESOURCES_DIR', 'TEST_NAME']),
                             sorted(list(steps[0]['variables'].keys())))
            self.assertEqual('echo', list(steps[1]['step'].keys())[0])
            self.assertEqual(sorted(['CURRENT_DIR', 'RESOURCES_DIR', 'TEST_NAME', 'user']),
                             sorted(list(steps[1]['variables'].keys())))
            self.assertEqual('val', steps[1]['variables']['user'])

    def test_run_multiple_tests(self):
        self.populate_file('first.yaml', '''---
                            steps:
                                - echo: {from: '1', register: {test: '{{ OUTPUT }}'}}
                            ''')
        self.populate_file('second.yaml', '''---
                           steps:
                               - echo: {from: '2', register: {test: '{{ OUTPUT }}'}}
                           ''')
        runner = Runner(self.test_dir, self.test_dir, None, output_format='json')
        self.assertTrue(runner.run_tests())
        reports = [f for f in listdir(join(self.test_dir, 'reports'))
                   if isfile(join(self.test_dir, 'reports', f)) and f.startswith('report')]
        self.assertEqual(1, len(reports))
        with open(join(self.test_dir, 'reports', reports[0]), 'r') as fp:
            obj = sorted(json.load(fp), key=lambda r: r['file'])
            self.assertEqual(2, len(obj))
            report = obj[0]
            self.assertEqual(join(self.test_dir, 'first.yaml'), report['file'])
            output = report['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual(sorted(['CURRENT_DIR', 'RESOURCES_DIR', 'TEST_NAME']),
                             sorted(list(steps[0]['variables'].keys())))
            self.assertEqual(1, steps[1]['variables']['test'])
            report = obj[1]
            self.assertEqual(join(self.test_dir, 'second.yaml'), report['file'])
            output = report['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual(sorted(['CURRENT_DIR', 'RESOURCES_DIR', 'TEST_NAME']),
                             sorted(list(steps[0]['variables'].keys())))
            self.assertEqual(2, steps[1]['variables']['test'])

    def test_run_with_include(self):
        self.populate_file('main.yaml', '''---
                include: simple_file.yaml
                steps:
                    - echo: 'hello'
                ''')
        self.populate_file('simple_file.yaml', '''---
                variables:
                    foo: bar
                steps:
                    - echo: {from: '{{ foo }}', to: foo.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, output_format='json')
        self.assertTrue(runner.run_tests())
        reports = [f for f in listdir(join(self.test_dir, 'reports'))
                   if isfile(join(self.test_dir, 'reports', f)) and f.startswith('report')]
        self.assertEqual(1, len(reports))
        with open(join(self.test_dir, 'reports', reports[0]), 'r') as fp:
            obj = json.load(fp)
            self.assertEqual(2, len(obj))
            report = obj[0]
            self.assertEqual('include', report['type'])
            self.assertEqual(join(self.test_dir, 'simple_file.yaml'), report['file'])
            output = report['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual(2, len(steps))
            self.assertEqual('OK', report['status'])
            report = obj[1]
            self.assertEqual('test', report['type'])
            self.assertEqual(join(self.test_dir, 'main.yaml'), report['file'])
            output = report['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual(2, len(steps))
            self.assertEqual('OK', report['status'])

    def test_run_multiple_includes(self):
        self.populate_file('main.yaml', '''---
                        include: 
                            - one.yaml
                            - two.yaml
                        steps:
                            - echo: 'hello'
                        ''')
        self.populate_file('one.yaml', '''---
                        variables:
                            foo: bar
                        steps:
                            - echo: {from: '{{ foo }}', to: one.output}
                        ''')
        self.populate_file('two.yaml', '''---
                            variables:
                                foo: bar
                            steps:
                                - echo: {from: '{{ foo }}', to: two.output}
                            ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, output_format='json')
        self.assertTrue(runner.run_tests())
        reports = [f for f in listdir(join(self.test_dir, 'reports'))
                   if isfile(join(self.test_dir, 'reports', f)) and f.startswith('report')]
        self.assertEqual(1, len(reports))
        with open(join(self.test_dir, 'reports', reports[0]), 'r') as fp:
            obj = json.load(fp)
            self.assertEqual(3, len(obj))
            report = obj[0]
            self.assertEqual('include', report['type'])
            self.assertEqual(join(self.test_dir, 'one.yaml'), report['file'])
            report = obj[1]
            self.assertEqual('include', report['type'])
            self.assertEqual(join(self.test_dir, 'two.yaml'), report['file'])
            report = obj[2]
            self.assertEqual('test', report['type'])
            self.assertEqual(join(self.test_dir, 'main.yaml'), report['file'])

    def test_run_with_include_named(self):
        self.populate_file('main.yaml', '''---
                include: 
                    file: simple_file.yaml
                    as: simple
                    run_on_include: false
                steps:
                    - echo: 'hello'
                    - run: 'simple'
                    - echo: 'world'
                ''')
        self.populate_file('simple_file.yaml', '''---
                variables:
                    foo: bar
                steps:
                    - echo: {from: '{{ foo }}', to: foo.output}
                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, output_format='json')
        self.assertTrue(runner.run_tests())
        reports = [f for f in listdir(join(self.test_dir, 'reports'))
                   if isfile(join(self.test_dir, 'reports', f)) and f.startswith('report')]
        self.assertEqual(1, len(reports))
        with open(join(self.test_dir, 'reports', reports[0]), 'r') as fp:
            obj = json.load(fp)
            self.assertEqual(1, len(obj))
            output = obj[0]['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual('echo', list(steps[0]['step'].keys())[0])
            self.assertEqual('echo', list(steps[1]['step'].keys())[0])
            self.assertEqual('run', list(steps[2]['step'].keys())[0])
            self.assertEqual('echo', list(steps[3]['step'].keys())[0])
            self.assertEqual('echo', list(steps[4]['step'].keys())[0])
            self.assertEqual('run', list(steps[5]['step'].keys())[0])
            self.assertEqual('echo', list(steps[6]['step'].keys())[0])
            self.assertEqual('echo', list(steps[7]['step'].keys())[0])

    def test_recursive_render_dict(self):
        self.populate_file('main.yaml', '''---
                        steps:
                          - wait:                                                                                                                                                                                                                
                                name: 'Waiting for postgres'                                                                                                                                                                                               
                                seconds: 30                                                                                                                                                                                                                
                                for:                                                                                                                                                                                                                 
                                  echo: {from: '1', to: foo.output}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, output_format='json')
        self.assertTrue(runner.run_tests())
        reports = [f for f in listdir(join(self.test_dir, 'reports'))
                   if isfile(join(self.test_dir, 'reports', f)) and f.startswith('report')]
        self.assertEqual(1, len(reports))
        with open(join(self.test_dir, 'reports', reports[0]), 'r') as fp:
            obj = json.load(fp)
            self.assertEqual(1, len(obj))
            output = obj[0]['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual('wait', list(steps[0]['step'].keys())[0])
            self.assertEqual('echo', list(steps[0]['step']['wait']['for'].keys())[0])

    def test_recursive_render_list(self):
        self.populate_file('main.yaml', '''---
                                steps:
                                  - wait:                                                                                                                                                                                                                
                                        name: 'Waiting for postgres'                                                                                                                                                                                               
                                        seconds: 30                                                                                                                                                                                                                
                                        for:                                                                                                                                                                                                                 
                                          - echo: {from: '1', to: foo.output}
                                          - echo: {from: '2', to: foo.output}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None, output_format='json')
        self.assertTrue(runner.run_tests())
        reports = [f for f in listdir(join(self.test_dir, 'reports'))
                   if isfile(join(self.test_dir, 'reports', f)) and f.startswith('report')]
        self.assertEqual(1, len(reports))
        with open(join(self.test_dir, 'reports', reports[0]), 'r') as fp:
            obj = json.load(fp)
            self.assertEqual(1, len(obj))
            output = obj[0]['output']
            steps = [o for o in output if 'step' in o]
            self.assertEqual('wait', list(steps[0]['step'].keys())[0])
            self.assertEqual('echo', list(steps[0]['step']['wait']['for'][0].keys())[0])
            self.assertEqual('echo', list(steps[0]['step']['wait']['for'][1].keys())[0])


    # py36 has dict insert ordering, while older implementations have some other.
    @staticmethod
    def _compare_operations(operation, *expectations):
        return any([operation == exp for exp in expectations])



import os
import unittest
from os.path import join
from shutil import copyfile

import test
from catcher.modules.log_storage import EmptyLogStorage
from catcher.utils import logger
from catcher.utils.file_utils import ensure_empty, remove_dir, ensure_dir
from catcher.utils.singleton import Singleton
from test import resources


class TestClass(unittest.TestCase):
    def __init__(self, test_name, method_name):
        super().__init__(method_name)
        self._test_name = test_name
        self._test_dir = test.get_test_dir(test_name)
        logger.configure('debug')
        logger.log_storage = EmptyLogStorage('empty')

    @property
    def test_name(self):
        return self._test_name

    @property
    def test_dir(self):
        return join(os.getcwd(), self._test_dir)

    @property
    def test_resources(self):  # test's resource dir (unique per testcase)
        return join(test.get_test_dir(self.test_name), 'resources')

    @property
    def global_resource_dir(self):  # global resource dir (part of project)
        return os.path.dirname(resources.__file__)

    def setUp(self):
        Singleton._instances = {}  # clean singleton between tests
        ensure_empty(test.get_test_dir(self.test_name))
        ensure_empty(self.test_resources)

    def tearDown(self):
        remove_dir(test.get_test_dir(self.test_name))

    def populate_file(self, file: str, content: str):
        with open(join(self.test_dir, file), 'w') as f:
            f.write(content)

    def populate_resource(self, file: str, content: str):
        filename = join(self.test_resources, file)
        ensure_dir(os.path.dirname(os.path.abspath(filename)))
        with open(filename, 'w') as f:
            f.write(content)

    def copy_resource(self, resource_name):
        """
        Copy resource from global resource to the resource dir
        """
        copyfile(join(self.global_resource_dir, resource_name), join(self.test_resources, resource_name))



import json
import os
import subprocess
from typing import Union, List, Optional, Tuple

from catcher.utils.logger import warning, debug
from catcher.utils import file_utils

from json import JSONDecodeError


def run_cmd(cmd: Union[List[str], str], variables, cwd=None, shell=False):
    env = _prepare_env(variables)
    process = subprocess.Popen(cmd,
                               cwd=cwd,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               universal_newlines=True,
                               env=env,
                               shell=shell)
    stdout, stderr = process.communicate()
    return process.returncode, _parse_output(stdout), stderr


def run_cmd_simple(cmd: str,
                   variables: dict,
                   env=None,
                   args: List[str] = None,
                   libraries=None) -> Union[dict, str]:
    """
    Run cmd with variables written in environment.
    :param args: cmd arguments
    :param cmd: to run
    :param variables: variables
    :param env: custom environment
    :param libraries: additional libraries used for source compilation
    :return: output in json (if can be parsed) or plaintext
    """
    env = _prepare_env(variables, env=env)
    cmd, cwd = _prepare_cmd(cmd, args, variables, libraries=libraries)
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env, cwd=cwd)
    if p.wait() == 0:
        out = p.stdout.read().decode()
        debug(out)
        return _parse_output(out)
    else:
        out = p.stdout.read().decode()
        warning(out)
        raise Exception('Execution failed.')


def _prepare_env(variables, env=None):
    if env is None:
        env = os.environ.copy()
    for k, v in variables.items():
        env[k] = str(v)
    return env


def _prepare_cmd(file: str, args: list = None, variables=None, libraries=None) -> Tuple[List[str], Optional[str]]:
    cmd = None
    cwd = None
    if file.endswith('.py'):  # python executable
        cmd = ['python', file]
    if file.endswith('.js'):  # node js executable
        cmd = ['node', file]
    # TODO Scala/Groovy compilation?
    if file.endswith('.java'):  # java source file (need to compile)
        classpath = _form_classpath_args(libraries)
        cmd = ['java'] + classpath + [_compile_java(file, variables, libraries=libraries)]
        cwd = variables['RESOURCES_DIR']  # compiled java class should be run from resources
    if file.endswith('.kt'):  # kotlin source file (need to compile)
        cmd = ['java', '-jar', _compile_kotlin(file, variables, libraries=libraries)]
        cwd = variables['RESOURCES_DIR']  # compiled java class should be run from resources
    if file.endswith('.jar'):  # executable jar
        cmd = ['java', '-jar', file]
    if cmd is None:  # local executable or other command
        cmd = [file]
    if args is not None:
        cmd += args
    debug(str(cmd))
    return cmd, cwd


def _compile_java(file, variables, libraries=None):
    resource_dir = variables['RESOURCES_DIR']
    classpath = _form_classpath(libraries)
    return_code, stdout, stderr = run_cmd('javac {} -d . *.java'.format(classpath),
                                          variables,
                                          cwd=resource_dir,
                                          shell=True)  # compile everything
    if return_code != 0:
        raise Exception("Can't compile {}. Out: {}, Err: {}".format(file, stdout, stderr))
    class_file = file_utils.find_resource(resource_dir, file_utils.get_filename(file), '.class')
    if len(class_file) > 1:
        debug('Found more than 1 resource for {}: {}. Use last.'.format(file_utils.get_filename(file), class_file))
        class_file = class_file[-1]
    else:
        class_file = class_file[0]
    module = file_utils.cut_part_path(resource_dir, class_file).replace('/', '.')
    return module.split('.class')[0]


def _compile_kotlin(file, variables, libraries=None):
    resource_dir = variables['RESOURCES_DIR']
    filename = file_utils.get_filename(file)
    return_code, stdout, stderr = run_cmd('kotlinc {}.kt -include-runtime -d {}.jar'.format(filename, filename),
                                          variables,
                                          cwd=resource_dir,
                                          shell=True)  # compile everything
    if return_code != 0:
        raise Exception("Can't compile {}. Out: {}, Err: {}".format(file, stdout, stderr))
    return filename + '.jar'


def _parse_output(output: str):
    try:
        return json.loads(output)
    except JSONDecodeError:
        return output


def _form_classpath(libraries) -> str:
    if libraries is None:
        return ''
    else:
        cp = '.:' + ':'.join(libraries)
        if cp.endswith('*'):
            return '-cp "{}"'.format(cp)
        return '-cp ' + cp


def _form_classpath_args(libraries) -> list:
    if libraries is None:
        return []
    else:
        return ['-cp'] + ['.:' + ':'.join(libraries)]



from typing import Optional, List

from catcher.core.include import Include
from catcher.core.test import Test
from catcher.utils.file_utils import read_source_file, get_filename, get_files


class ParseResult:
    def __init__(self, test_to_run, run_on_include=None, parse_error=None) -> None:
        super().__init__()
        self.run_on_include = run_on_include
        self.parse_error = parse_error
        self.test = test_to_run

    @property
    def should_run(self) -> bool:
        return self.parse_error is None


class Parser:

    def __init__(self, path: str, inventory_path: Optional[str]) -> None:
        self.path = path
        self.inventory = inventory_path

    def read_inventory(self) -> dict:
        if self.inventory is not None:
            inv_vars = read_source_file(self.inventory)
            inv_vars['INVENTORY'] = get_filename(self.inventory)
            inv_vars['INVENTORY_FILE'] = self.inventory
            return inv_vars
        return {}

    def read_tests(self, test_path: str) -> List[ParseResult]:
        """
        Parse all tests at path, return pairs of (tests run before the test, test)
        """
        results = []
        for test_file in get_files(test_path):
            try:
                raw_test = self.read_test(test_file)
                if raw_test.ignore:  # ignore the test - do not need to check imports (they may be missing)
                    results += [ParseResult(raw_test, run_on_include=[])]
                    continue
                run_on_include = self.fill_includes_recursive(test_file, raw_test, None)
                results += [ParseResult(raw_test, run_on_include=run_on_include)]
            except Exception as e:
                results += [ParseResult(test_file, parse_error=str(e))]
        return results

    def fill_includes_recursive(self, parent, raw_test, all_includes):
        run_on_include = []
        all_includes, raw_includes = Include.read_includes(self.path, parent, raw_test.includes, all_includes)
        raw_test.includes = {}  # clear & fill real includes
        for include in raw_includes:
            test = self.read_test(include.file)
            test.include = include  # remember include's properties (like ignore_errors)
            if include.alias is not None:
                raw_test.includes[include.alias] = test
            if include.run_on_include:
                run_on_include += [test]
            run_on_include += self.fill_includes_recursive(include.file, test, all_includes)
        return run_on_include

    def read_test(self, test_file: str):
        body = read_source_file(test_file)
        return Test(self.path,
                    test_file,
                    includes=body.get('include', []),
                    variables=body.get('variables', {}),
                    config=body.get('config', {}),
                    steps=body.get('steps', []),
                    final=body.get('finally', []),
                    ignore=body.get('ignore', False))



from typing import Union, Optional, List

from catcher.steps import step
from catcher.steps.external import External
from catcher.steps.step import Step
from catcher.utils.misc import merge_two_dicts
from catcher.utils.module_utils import prepare_modules
from catcher.utils.singleton import Singleton


class StepFactory(metaclass=Singleton):

    def __init__(self, modules: Optional[List[str]] = None) -> None:
        if modules is None:
            modules = []
        self.modules = merge_two_dicts(prepare_modules(modules, step.registered_steps), step.registered_steps)

    def get_actions(self, path: str, step_to_run: dict) -> [Step]:
        [action] = step_to_run.keys()
        body = step_to_run[action]
        steps = []
        if 'actions' in body:
            for action_step in body['actions']:
                steps.append(self.get_action(path, action, action_step))
        else:
            steps.append(self.get_action(path, action, body))
        return steps

    def get_action(self, path: str, action, body: dict or str) -> Step:
        module = self.modules.get(action, None)
        if module is not None:
            if isinstance(module, str):  # external module (some third party script)
                body = self.ensure_body_dict(body, path)
                body['_module'] = module
                return External(**body)
            else:  # python module
                body = self.ensure_body_dict(body, path)
                return module(**body)
        raise FileNotFoundError('Can\'t find module for action: ' + action)

    def ensure_body_dict(self, body: Union[str, dict], path) -> dict:
        def get_actions_fun(act):
            return self.get_actions(path, act)

        def get_action_fun(x):
            return self.get_action(path, x[0], x[1])

        if isinstance(body, str):
            body = {'_body': body}
        body['_get_actions'] = get_actions_fun
        body['_get_action'] = get_action_fun
        body['_path'] = path
        return body



from catcher.utils.singleton import Singleton
from catcher.modules.compose import DockerCompose
from catcher.modules.requirements import Requirements
from catcher.utils.logger import debug


class ModulesFactory(metaclass=Singleton):

    def __init__(self, **kwargs) -> None:
        """
        :param kwargs: resources_dir param must exist for the initial load
        """
        super().__init__()
        # TODO find all submodules of catcher.modules.Module dynamically!
        self._modules = {'compose': DockerCompose(kwargs['resources_dir']),
                         'requirements': Requirements(kwargs['resources_dir'])}
        debug('Loaded modules: {}'.format(list(self._modules.keys())))

    @property
    def modules(self):
        return self._modules



import os


def check_file(file: str, expected: str) -> bool:
    if not os.path.exists(file):
        print('file {} doesn\'t exist'.format(file))
        return False
    with open(file, 'r') as f:
        content = f.read()
        if content != expected:
            print('Got ' + content + ' instead of expected ' + expected)
            return False
        return True


def read_file(file: str) -> str:
    with open(file, 'r') as f:
        return f.read()



from os.path import join

from catcher.utils.file_utils import ensure_dir, get_files, read_source_file
from test.abs_test_class import TestClass


class UtilsTest(TestClass):
    def __init__(self, method_name):
        super().__init__('utils_test', method_name)

    @property
    def yaml_template(self) -> str:
        return '''---
        vars:
            foo: bar
        steps:
            - one
            - two
            - three
        '''

    @property
    def yaml_parsed(self) -> dict:
        return {'steps': ['one', 'two', 'three'], 'vars': {'foo': 'bar'}}

    # read and parse yaml file
    def test_read_yaml_simple(self):
        self.__populate_files(['test_file.yaml'])
        file = join(self.test_dir, 'test_file.yaml')
        result = read_source_file(file)
        self.assertEqual(self.yaml_parsed, result)

    # find file, when launched via `catcher /path/to/file.yaml`
    def test_find_one_file(self):
        self.__populate_files(['test_file.yaml'])
        file = join(self.test_dir, 'test_file.yaml')
        result = get_files(file)
        self.assertEqual([file], result)

    # find all files, when launched via `catcher /path/to`
    def test_find_all_in_dir(self):
        files = ['d', 'd/test1.yaml', 'd/test2.yaml']
        self.__populate_files(files)
        result = get_files(join(self.test_dir, 'd'))
        expected = [join(self.test_dir, f) for f in files if f.endswith('yaml')]
        expected.reverse()
        self.assertEqual(sorted(expected), sorted(result))

    # find all files, when launched via `catcher /path/to` with subdirs
    def test_find_all_with_subdir(self):
        files = ['d', 'd/d1', 'd/test1.yaml', 'd/test2.yaml', 'd/d1/test3.yaml']
        self.__populate_files(files)
        result = get_files(join(self.test_dir, 'd'))
        expected = [join(self.test_dir, f) for f in files if f.endswith('yaml')]
        expected.reverse()
        self.assertEqual(sorted(expected), sorted(result))

    # find all files, when launched via `catcher /path/to` with subdirs and other files
    def test_find_all_filter_unknown(self):
        files = ['d', 'd/d1', 'd/test1.yaml', 'd/test2.yaml', 'd/d1/test3.yaml']
        self.__populate_files(files)
        with open(join(self.test_dir, 'd/Readme.md'), 'w') as f:
            f.write('first_readme')
        with open(join(self.test_dir, 'd/d1/Readme.md'), 'w') as f:
            f.write('second_readme')
        result = get_files(join(self.test_dir, 'd'))
        expected = [join(self.test_dir, f) for f in files if f.endswith('yaml')]
        expected.reverse()
        self.assertEqual(sorted(expected), sorted(result))

    def __populate_files(self, files):
        for file in files:
            if file.endswith('.yaml'):
                with open(join(self.test_dir, file), 'w') as f:
                    f.write(str(self.yaml_template))
            else:
                ensure_dir(join(self.test_dir, file))



import json
from os.path import join

import yaml

from catcher.core.runner import Runner
from test.abs_test_class import TestClass


class InputsTest(TestClass):
    def __init__(self, method_name):
        super().__init__('inputs_test', method_name)

    @property
    def inventory(self) -> dict:
        return {'list': ['a', 'b', 'c']}

    @property
    def test(self) -> dict:
        return {
            'variables': {'dict': {'a': 1, 'b': 2, 'c': 3}},
            'steps': [
                {'check': {'contains': {'the': 'a', 'in': '{{ list }}'}}},
                {'check': {'contains': {'the': 'a', 'in': '{{ dict }}'}}}
            ]
        }

    def test_yaml_lowercase(self):
        with open(join(self.test_dir, 'inventory.yaml'), 'w') as outfile:
            yaml.dump(self.inventory, outfile)
        with open(join(self.test_dir, 'main.yaml'), 'w') as outfile:
            yaml.dump(self.test, outfile)
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), join(self.test_dir, 'inventory.yaml'))
        self.assertTrue(runner.run_tests())

    def test_yaml_uppercase(self):
        with open(join(self.test_dir, 'inventory.YAML'), 'w') as outfile:
            yaml.dump(self.inventory, outfile)
        with open(join(self.test_dir, 'main.YAML'), 'w') as outfile:
            yaml.dump(self.test, outfile)
        runner = Runner(self.test_dir, join(self.test_dir, 'main.YAML'), join(self.test_dir, 'inventory.YAML'))
        self.assertTrue(runner.run_tests())

    def test_yml_lowercase(self):
        with open(join(self.test_dir, 'inventory.yml'), 'w') as outfile:
            yaml.dump(self.inventory, outfile)
        with open(join(self.test_dir, 'main.yml'), 'w') as outfile:
            yaml.dump(self.test, outfile)
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yml'), join(self.test_dir, 'inventory.yml'))
        self.assertTrue(runner.run_tests())

    def test_yml_uppercase(self):
        with open(join(self.test_dir, 'inventory.YML'), 'w') as outfile:
            yaml.dump(self.inventory, outfile)
        with open(join(self.test_dir, 'main.YML'), 'w') as outfile:
            yaml.dump(self.test, outfile)
        runner = Runner(self.test_dir, join(self.test_dir, 'main.YML'), join(self.test_dir, 'inventory.YML'))
        self.assertTrue(runner.run_tests())

    def test_json_lowercase(self):
        with open(join(self.test_dir, 'inventory.json'), 'w') as outfile:
            json.dump(self.inventory, outfile)
        with open(join(self.test_dir, 'main.json'), 'w') as outfile:
            json.dump(self.test, outfile)
        runner = Runner(self.test_dir, join(self.test_dir, 'main.json'), join(self.test_dir, 'inventory.json'))
        self.assertTrue(runner.run_tests())

    def test_json_uppercase(self):
        with open(join(self.test_dir, 'inventory.JSON'), 'w') as outfile:
            json.dump(self.inventory, outfile)
        with open(join(self.test_dir, 'main.JSON'), 'w') as outfile:
            json.dump(self.test, outfile)
        runner = Runner(self.test_dir, join(self.test_dir, 'main.JSON'), join(self.test_dir, 'inventory.JSON'))
        self.assertTrue(runner.run_tests())



import collections
import itertools
import json

from catcher.steps.check import Operator
from catcher.steps.step import Step, update_variables
from catcher.utils.logger import debug
from catcher.utils.misc import fill_template_str, try_get_objects


class Loop(Step):
    """
    Repeat one or several actions till the condition is true or for each element of the collection. Is useful, when you
    need to wait for some process to start or for async execution to finish.

    :Input:

    :while: perform action while the condition is true

    - if: your condition. It can be in short format: `if: '{{ counter < 10 }}'` and
            long one: `if: {equals: {the: '{{ counter }}', is_not: 10000}}`. The clause
            format is the same as in [checks](checks.md)
    - do: the aciton to be performed. Can be a list of actions or single one.
    - max_cycle: the limit of reductions. *Optional* default is no limit.

    :foreach: iterate data structure

    - in: variable or static list. **ITEM** variable can be used to access each element of the data structure.
            Data structure can be list, dict or any other python data structure which supports iteration.
    - do: the action to be performed. Can be a list of actions or single one.

    :Examples:

    Perform a single echo wile counter is less than 10
    ::

        loop:
            while:
                if: '{{ counter < 10 }}'
                do:
                    echo: {from: '{{ counter + 1 }}', register: {counter: '{{ OUTPUT }}'}}
                max_cycle: 100000

    Perform to actions: consume message from kafka and send token via POST http. Do it until server
    returns passed true in http response.
    ::

        loop:
            while:
                if:
                    equals: {the: '{{ passed }}', is_not: True}
                do:
                    - kafka:
                          consume:
                              server: '127.0.0.1:9092'
                              group_id: 'test'
                              topic: 'test_consume_with_timestamp'
                              timeout: {seconds: 5}
                              where:
                                  equals: '{{ MESSAGE.timestamp > 1000 }}'
                          register: {token: '{{ OUTPUT.data.token }}'}
                    - http:
                        post:
                          headers: {Content-Type: 'application/json'}
                          url: 'http://test.com/check_my_token'
                          body: {'token': '{{ token }}'}
                        register: {passed: '{{ OUTPUT.passed }}'}

    Iterate over `iterator` variable, produce each element to kafka as json and debug it to file.
    ::

        loop:
            foreach:
                in: '{{ iterator }}'
                do:
                    - kafka:
                          produce:
                              server: '127.0.0.1:9092'
                              topic: 'test_produce_json'
                              data: '{{ ITEM|tojson }}'
                    - echo: {from: '{{ ITEM }}', to: '{{ ITEM["filename"] }}.output'}

    Iterate over several different configurations.
    ::
        variables:
            db_1: 'test:test@localhost:5433/db1'
            db_2:
                url: 'test:test@localhost:5434/db2'
                type: postgres
            db_3:
                dbname: 'db3'
                user: 'test'
                password: 'test'
                host: 'localhost'
                port: 5435
                type: 'postgres'
        steps:
            - loop:
                foreach:
                    in: '["{{ db_1 }}", {{ db_2 }}, {{ db_3 }}]'
                    do:
                        - postgres:
                            request:
                                conf: '{{ ITEM }}'
                                query: 'select count(*) from test'
                            register: {documents: '{{ OUTPUT }}'}
                        - check:
                            equals: {the: '{{ documents.count }}', is: 2}


    Note that db_1 template has additional quotes ``"{{ db_1 }}"``. Your **in** should contain valid object. As db_1 is
    just a string - it should be put in quotes. Otherwise **in** value will be corrupted.
    Always make sure your in value is valid. It may also have some difficulties with json as string.
    """

    def __init__(self, _get_action=None, _get_actions=None, **kwargs):
        super().__init__(**kwargs)
        self.type = Step.filter_predefined_keys(kwargs)  # while/foreach
        do = kwargs[self.type]['do']
        if len(do) == 1:  # just action
            if isinstance(do, list):  # list with single action
                do = do[0]
            [loop_action] = do.keys()
            self.do_action = [_get_action((loop_action, do[loop_action]))]
        else:
            self.do_action = list(itertools.chain.from_iterable([_get_actions(act) for act in do]))
        self.max_cycle = kwargs[self.type].get('max_cycle')
        if self.type == 'while':
            if_clause = kwargs['while']['if']
            if isinstance(if_clause, str):
                self.if_clause = {'equals': if_clause}
            else:
                self.if_clause = if_clause
        elif self.type == 'foreach':
            self.in_var = kwargs['foreach']['in']
        else:
            raise ValueError('Wrong configuration for step: ' + str(kwargs))

    @update_variables
    def action(self, includes: dict, variables: dict) -> dict:
        output = variables
        if self.type == 'while':
            operator = Operator.find_operator(self.if_clause)
            while operator.operation(output):
                output = self.__run_actions(includes, output)
                if self.max_cycle is not None:
                    if self.max_cycle == 0:
                        break
                    self.max_cycle = self.max_cycle - 1
            return output
        elif self.type == 'foreach':
            loop_var = try_get_objects(fill_template_str(json.dumps(self.in_var), variables))
            if not isinstance(loop_var, collections.Iterable):
                raise ValueError(str(loop_var) + ' is not iterable')
            for entry in loop_var:
                debug('Looping over {}'.format(entry))
                output['ITEM'] = entry
                output = self.__run_actions(includes, output)
        return output

    def __run_actions(self, includes, variables: dict) -> dict:
        output = variables
        for action in self.do_action:
            try:
                output = action.action(includes, output)
            except Exception as e:
                if action.ignore_errors:
                    debug('{} got {} but we ignore it'.format(fill_template_str(action.name, variables), e))
                    break
                raise e
        return output



import inspect
import io
import json
import ntpath
import os
import shutil
from glob import glob
from os.path import join
from typing import List

import yaml


def get_module_filename(module) -> str:
    return get_filename(inspect.getfile(module))


def get_filename(filename: str) -> str:
    return ntpath.basename(filename).split('.')[0]


def cut_path(tests_path, test_path):
    return get_filename(cut_part_path(tests_path, test_path))


def cut_part_path(tests_path, test_path):
    if tests_path == test_path:
        return get_filename(test_path)
    # quick fix: fail with include step: can't mix absolute and relative path
    # TODO need to be rewritten
    try:
        common = os.path.commonpath([test_path, tests_path])
    except ValueError:
        return get_filename(test_path)
    return test_path.split(common)[1][1:]


# Get list of yaml files in dir and subdirs
def get_files(path: str) -> list:
    if not os.path.exists(path):
        raise FileNotFoundError('No such path: ' + path)
    file = []
    if os.path.isdir(path):
        for f in os.listdir(path):
            path_in_dir = join(path, f)
            if os.path.isfile(path_in_dir) and (path_in_dir.endswith('yaml') or path_in_dir.endswith('yml')):
                file.append(path_in_dir)
            elif os.path.isdir(path_in_dir):
                file += get_files(path_in_dir)
    else:
        file.append(path)
    return file


def find_resource(path: str, resource_name: str, extension=".*") -> List[str]:
    files = []
    pattern = resource_name + extension
    for d, _, _ in os.walk(path):
        files.extend(glob(os.path.join(d, pattern)))
    if not files:
        raise Exception('No resource found for {}'.format(resource_name))
    return files


def read_source_file(file: str) -> dict:
    if not os.path.exists(file):
        raise FileNotFoundError('No such file: ' + file)
    if file.lower().endswith('json'):
        return _read_json_file(file)
    else:
        return _read_yaml_file(file)


def read_file(file: str) -> str:
    if not os.path.exists(file):
        raise FileNotFoundError('No such file: ' + file)
    with io.open(file, mode='r', encoding='utf-8') as stream:
        return stream.read()


# If dir exists delete and and create again
def ensure_empty(path: str):
    remove_dir(path)
    ensure_dir(path)


def remove_dir(path: str):
    if os.path.exists(path):
        shutil.rmtree(path)


def ensure_dir(path: str):
    if not os.path.exists(path):
        os.makedirs(path)


def _read_yaml_file(file: str) -> dict:
    with open(file, 'r') as stream:
        try:
            return yaml.load(stream, Loader=yaml.FullLoader) or {}
        except yaml.YAMLError as exc:
            raise yaml.YAMLError('Wrong YAML format for file ' + file + ' : ' + str(exc))


def _read_json_file(file: str) -> dict:
    with open(file, 'r') as stream:
        try:
            return json.load(stream) or {}
        except yaml.YAMLError as exc:
            raise yaml.YAMLError('Wrong YAML format for file ' + file + ' : ' + str(exc))



import grpc
from concurrent import futures

import calculator_pb2
import calculator_pb2_grpc

import math


class CalculatorServicer(calculator_pb2_grpc.CalculatorServicer):

    def SquareRoot(self, request, context):
        response = calculator_pb2.Number()
        response.value = math.sqrt(request.value)
        return response


server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
calculator_pb2_grpc.add_CalculatorServicer_to_server(CalculatorServicer(), server)
server.add_insecure_port('[::]:50051')



from os.path import isfile, join

from catcher.modules import Module
from catcher.utils.logger import info
from catcher.utils.module_utils import is_package_installed


class DockerCompose(Module):
    """
    :Docker-compose module:

    Will automatically run `docker-compose up -d` before your test and `down` after it if this
    module was enabled and you have `docker-compose.yml` file in your directory.

    :Enable this module:

    - run `pip install catcher[compose]`. It will installs all requirements, if they were not installed.

    """
    def __init__(self, resources_dir) -> None:
        super().__init__(resources_dir)
        self._cmd = None
        self._options = {'--detach': True,
                         'SERVICE': "",
                         '--no-deps': False,
                         '--abort-on-container-exit': False,
                         '--no-recreate': True,
                         '--remove-orphans': False,
                         '--always-recreate-deps': False,
                         '--force-recreate': False,
                         '--build': False,
                         '--no-build': False,
                         '--no-color': False,
                         '--rmi': 'none',
                         '--volumes': "",
                         '--follow': False,
                         '--timestamps': False,
                         '--scale': {}
                         }

    def before(self, *args, **kwargs):
        """
        Will run `docker-compose up -d` only in case docker-compose is installed locally
        and there is a docker-compose.yml file in resources directory
        """
        compose = self.find_resource_file()
        if compose and is_package_installed('compose'):
            from compose.cli.main import TopLevelCommand, project_from_options
            info('Starting docker-compose. Please wait.')
            self._options['-f'] = join(self._resources, compose),
            self._cmd = TopLevelCommand(project_from_options(self._resources, self._options))
            self._cmd.up(self._options)

    def after(self, *args, **kwargs):
        if self._cmd:
            self._cmd.down(self._options)

    def filter_resource(self, file: str) -> bool:
        # use only default docker-compose.yml for now, will change it in future if needed.
        return isfile(join(self._resources, file)) and file == 'docker-compose.yml'



import os
from os.path import join

from catcher.steps.step import Step, update_variables
from catcher.utils.file_utils import read_file
from catcher.utils.logger import info
from catcher.utils.misc import fill_template, fill_template_str
from catcher.utils import file_utils, module_utils


class Echo(Step):
    """
    Print a string constant, variable or file to the console or file.

    :Input:

    :from: data source. Can be variable or constant string
    :from_file: file in resources.
    :to: output to file. *Optional* If not set - stdout will be used. **Not** resources-related

    Has short from which just prints variable to stdout.

    :Examples:

    Use short form to print variable to stdout
    ::

        echo: '{{ var }}'

    Print constant + variable to file
    ::

        echo: {from: 'constant and {{ var }}', to: debug.output}

    Use echo to register new variable
    ::

        echo: {from: '{{ RANDOM_STR }}@test.com', register: {user_email: '{{ OUTPUT }}'}}

    Read file content to a variable
    ::

        echo: {from_file: debug.output, to: '{{ user_email }}'}

    """

    def __init__(self, _path: str = None, _body=None, to=None, from_file=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.source = _body if _body else kwargs.get('from')
        self.source_file = from_file
        self.dst = to
        self.path = _path
        if self.source is None and self.source_file is None:
            raise ValueError('Incorrect arguments for echo.')

    @update_variables
    def action(self, includes: dict, variables: dict) -> tuple:
        if self.source_file:  # read from file
            out = self._read_file(variables)
        else:
            out = fill_template(self.source, variables)
        if self.dst is None:
            info(out)
        else:
            dst = fill_template(self.dst, variables)
            path = fill_template(self.path, variables)
            filename = join(path, dst)
            file_utils.ensure_dir(os.path.dirname(os.path.abspath(filename)))
            with open(filename, 'w') as f:
                f.write(str(out))
        return variables, out

    def _read_file(self, variables):
        resources = variables['RESOURCES_DIR']
        file = join(resources, fill_template_str(self.source_file, variables))
        if self.source_file.endswith('.py'):
            try:
                read = module_utils.load_external_actions(file)
                if read is None:
                    raise Exception("Can't parse file")
                external = [f for f in dir(read) if not f.startswith('__')]  # load not internal values as output
                return dict([(n, getattr(read, n)) for n in external])
            except Exception:  # if we can't parse python file - return it as string (old normal way)
                return fill_template_str(read_file(file), variables)
        else:
            return fill_template_str(read_file(file), variables)



from os.path import dirname, join

from catcher.steps.step import Step, update_variables
from catcher.utils import module_utils
from catcher.utils import file_utils
from catcher.utils.misc import fill_template, fill_template_str, try_get_objects


class GRPC(Step):
    """
    Perform a remote procedure call with protobuffers layer.

    :Input:

    :call: Make a remote procedure call

    - url: server url
    - function: service and method you are going to call separated by dot. Case insensitive (MyClass.my_function)
    - schema: path to the .proto resource file. *Optional* Ignore it if reflection is configured on the
      server side
    - data: data to pass. *Optional*

    :Examples:
    calculator.proto
    ::

        message Number {
            float value = 1;
        }

        service Calculator {
            rpc SquareRoot(Number) returns (Number) {}
        }

    test
    ::

        grpc:
            call:
                url: 'localhost:50051'
                function: calculator.squareroot
                schema: 'calculator.proto'
                data: {'value': 2}
            register: {'my_value': '{{ OUTPUT.value }}'

    Complex schema case::

        grpc:
            call:
                url: 'localhost:50051'
                function: greeter.greet
                schema: 'greeter.proto'
                data:
                    result:
                        url: '{{ my_url }}'
                        title: 'test'
                        snippets: 'test2'
            register: {value: '{{ OUTPUT.name }}'}

    Useful tip: if you'd like to use templates in your .proto file - do not do it in the original resources, as Catcher
    shouldn't modify them. Use echo step to fill a template and create another .proto file for you.
    """

    def __init__(self, call=None, **kwargs) -> None:
        super().__init__(**kwargs)
        if call:
            [service, method] = call['function'].split('.')
            self.service = service.lower()
            self.url = call['url']
            self.method = method.lower()
            self.schema = call.get('schema')
            self.data = call.get('data', {})

    @update_variables
    def action(self, includes: dict, variables: dict) -> dict or tuple:
        import grpc
        channel = grpc.insecure_channel(fill_template(self.url, variables))
        if self.schema:
            self._compile_proto_files(variables)
            client = self._open_channel(channel, variables)
            method, input_arg = self._compose_arg(variables)
            return variables, getattr(client, method)(input_arg)
        else:
            raise Exception('Reflection not supported (yet)')

    def _compile_proto_files(self, variables):
        """
        Compile .proto resource into the definition (_pb2) and client code (_pb2_grpc)
        """
        from grpc.tools import command
        schema = fill_template(self.schema, variables)
        command.build_package_protos(join(variables['RESOURCES_DIR'], dirname(schema)), strict_mode=True)

    def _open_channel(self, channel, variables):
        """
        Search for stub in generated module _pb2_grpc. Instantiate it with channel and return.
        """
        schema = fill_template(self.schema, variables)
        mod = module_utils.load_external_actions(join(variables['RESOURCES_DIR'],
                                                      file_utils.get_filename(schema) + '_pb2_grpc.py'))
        classes = module_utils.get_all_classes(mod)
        classes = {k.lower(): v for k, v in classes.items()}
        stub = classes.get(fill_template(self.service, variables) + 'stub')
        if not stub:
            raise Exception('Can\'t find stub in generated code. Something went wrong')
        return stub(channel)

    def _compose_arg(self, variables):
        """
        Search for the method in generated _pb2. Resolve input type by the method's input.
        """
        schema = fill_template(self.schema, variables)
        service = fill_template(self.service, variables)
        method = fill_template(self.method, variables)
        # search for the service
        mod = module_utils.load_external_actions(join(variables['RESOURCES_DIR'],
                                                      file_utils.get_filename(schema) + '_pb2.py'))
        services = {str(k).lower(): v for k, v in mod.DESCRIPTOR.services_by_name.items()}
        service = services.get(service)
        if not service:
            raise Exception('Unable to find service {} in {}'.format(service,
                                                                     file_utils.get_filename(schema) + '_pb2'))
        # find service's method
        methods = dict([(f.name.lower(), f) for f in service.methods])
        method = methods.get(method)
        if not method:
            raise Exception('No method {} in service {}'.format(method, service))
        # find method's input type
        input_type = method.input_type
        classes = module_utils.get_all_classes(mod)
        data = try_get_objects(fill_template_str(self.data, variables))
        return method.name, classes[input_type.name](**data)



from os.path import join

from catcher.utils.module_utils import load_external_actions

TEST_DIR = 'test/tmp'

load_external_actions('catcher.steps')


def get_test_dir(testcase: str) -> str:
    return join(TEST_DIR, testcase)



import os
from os.path import join

from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from test.test_utils import check_file


class FinallyTest(TestClass):
    def __init__(self, method_name):
        super().__init__('finally_test', method_name)

    def test_run_finally(self):
        self.populate_file('main.yaml', '''---
        steps:
            - echo: {from: '123', to: sys_env.output}
        finally:
            - sh: 
                command: 'rm sys_env.output'
                path: '{{ CURRENT_DIR }}'
        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        None,
                        system_environment=dict(os.environ))
        runner.run_tests()
        self.assertFalse(check_file(join(self.test_dir, 'sys_env.output'), '123'))

    def test_run_finally_fail_fail(self):
        self.populate_file('main.yaml', '''---
                        steps:
                            - check: {equals: {the: '1', is: '2'}}
                        finally:
                            - sh: 
                                command: 'mkdir test'
                                path: '{{ CURRENT_DIR }}'
                                run_if: 'fail'
                        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        None,
                        system_environment=dict(os.environ))
        runner.run_tests()
        self.assertTrue(os.path.exists(join(self.test_dir, 'test')))

    def test_run_finally_fail_pass(self):
        self.populate_file('main.yaml', '''---
                        steps:
                            - check: {equals: {the: '2', is: '2'}}
                        finally:
                            - sh: 
                                command: 'mkdir test'
                                path: '{{ CURRENT_DIR }}'
                                run_if: 'fail'
                        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        None,
                        system_environment=dict(os.environ))
        runner.run_tests()
        self.assertFalse(os.path.exists(join(self.test_dir, 'test')))

    def test_run_finally_pass_fail(self):
        self.populate_file('main.yaml', '''---
                        steps:
                            - check: {equals: {the: '1', is: '2'}}
                        finally:
                            - sh: 
                                command: 'mkdir test'
                                path: '{{ CURRENT_DIR }}'
                                run_if: 'pass'
                        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        None,
                        system_environment=dict(os.environ))
        runner.run_tests()
        self.assertFalse(os.path.exists(join(self.test_dir, 'test')))

    def test_run_finally_pass_pass(self):
        self.populate_file('main.yaml', '''---
                        steps:
                            - check: {equals: {the: '2', is: '2'}}
                        finally:
                            - sh: 
                                command: 'mkdir test'
                                path: '{{ CURRENT_DIR }}'
                                run_if: 'pass'
                        ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        None,
                        system_environment=dict(os.environ))
        runner.run_tests()
        self.assertTrue(os.path.exists(join(self.test_dir, 'test')))

    def test_run_finally_always_pass(self):
        self.populate_file('main.yaml', '''---
                                steps:
                                    - check: {equals: {the: '2', is: '2'}}
                                finally:
                                    - sh: 
                                        command: 'mkdir test'
                                        path: '{{ CURRENT_DIR }}'
                                        run_if: 'always'
                                ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        None,
                        system_environment=dict(os.environ))
        runner.run_tests()
        self.assertTrue(os.path.exists(join(self.test_dir, 'test')))

    def test_run_finally_always_fail(self):
        self.populate_file('main.yaml', '''---
                                steps:
                                    - check: {equals: {the: '1', is: '2'}}
                                finally:
                                    - sh: 
                                        command: 'mkdir test'
                                        path: '{{ CURRENT_DIR }}'
                                        run_if: 'always'
                                ''')
        runner = Runner(self.test_dir,
                        join(self.test_dir, 'main.yaml'),
                        None,
                        system_environment=dict(os.environ))
        runner.run_tests()
        self.assertTrue(os.path.exists(join(self.test_dir, 'test')))



from catcher.steps.step import Step, update_variables
from catcher.utils.misc import fill_template
from catcher.utils.logger import debug
from catcher.utils import external_utils


class Sh(Step):
    """
    Run shell command and return output.

    :Input:

    - command: Command to run.
    - path: Path to be used as a root for the command. *Optional*.
    - return_code: expected return code. *Optional*. 0 is default.

    :Examples:

    List current directory
    ::

        - sh:
            command: 'ls -la'

    Determine if running in docker
    ::

        variables:
            docker: true
        steps:
            - sh:
                command: "grep 'docker|lxc' /proc/1/cgroup"
                return_code: 1
                ignore_errors: true
                register: {docker: false}
            - echo: {from: 'In docker: {{ docker }}'}

    """

    def __init__(self, command=None, path=None, return_code=0, **kwargs) -> None:
        super().__init__(**kwargs)
        self._cmd = command
        self._path = path
        self._return_code = return_code

    @update_variables
    def action(self, includes: dict, variables: dict) -> dict or tuple:
        cmd = fill_template(self._cmd, variables)
        return_code, stdout, stderr = external_utils.run_cmd(cmd.split(' '),
                                                             variables,
                                                             fill_template(self._path, variables))
        if return_code != int(fill_template(self._return_code, variables)):
            debug('Process return code {}.\nStderr is {}\nStdout is {}'.format(return_code, stderr, stdout))
            raise Exception(stderr)
        return variables, stdout



from os.path import join

import os

from catcher.core.runner import Runner
from test.abs_test_class import TestClass
from test.test_utils import check_file


class RunRegisteredIncludeTest(TestClass):
    def __init__(self, method_name):
        super().__init__('run_registered_includes_test', method_name)

    # test register include and run later
    def test_simple_include(self):
        self.populate_file('main.yaml', '''---
        include: 
            file: simple_file.yaml
            as: simple
            run_on_include: false
        steps:
            - run: 'simple'
        ''')
        self.populate_file('simple_file.yaml', '''---
        variables:
            foo: bar
        steps:
            - echo: {from: '{{ foo }}', to: foo.output}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'foo.output'), 'bar'))

    # test register include and run only selected tag
    def test_register_run_tags(self):
        self.populate_file('main.yaml', '''---
        include: 
            file: simple_file.yaml
            as: simple
        steps:
            - run: 'simple.one'
        ''')
        self.populate_file('simple_file.yaml', '''---
        variables:
            foo: 1
            baz: 2
            bar: 3
        steps:
            - echo: {from: '{{ foo }}', to: foo.output, tag: one}
            - echo: {from: '{{ baz }}', to: baz.output}
            - echo: {from: '{{ bar }}', to: bar.output, tag: one}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'foo.output'), '1'))
        self.assertTrue(not os.path.exists(join(self.test_dir, 'baz.output')))
        self.assertTrue(check_file(join(self.test_dir, 'bar.output'), '3'))

    # test register include and run only selected tag (long form), include name with dot
    def test_register_run_tags_with_dot(self):
        self.populate_file('main.yaml', '''---
        include: 
            file: simple_file.yaml
            as: simple.with_dot
        steps:
            - run: 
                include: 'simple.with_dot.one'
        ''')
        self.populate_file('simple_file.yaml', '''---
        variables:
            foo: 1
            baz: 2
            bar: 3
        steps:
            - echo: {from: '{{ foo }}', to: foo.output, tag: one}
            - echo: {from: '{{ baz }}', to: baz.output}
            - echo: {from: '{{ bar }}', to: bar.output, tag: one}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'foo.output'), '1'))
        self.assertTrue(not os.path.exists(join(self.test_dir, 'baz.output')))
        self.assertTrue(check_file(join(self.test_dir, 'bar.output'), '3'))

    # test include can be run from variable
    def test_run_from_template(self):
        self.populate_file('main.yaml', '''---
        include: 
            - file: one.yaml
              as: one
            - file: two.yaml
              as: two
            - file: determine_include.yaml
              as: determine_include
        steps:
            - run: 'determine_include'
            - run: '{{ include }}'
        ''')
        self.populate_file('one.yaml', '''---
        steps:
            - echo: {from: 'bar', to: foo.output}
        ''')
        self.populate_file('two.yaml', '''---
        steps:
            - echo: {from: 'baz', to: foo.output}
        ''')
        self.populate_file('determine_include.yaml', '''---
        steps:
            - echo: {from: 'hello', register: {include: one}}
        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'foo.output'), 'bar'))
        self.assertTrue(not os.path.exists(join(self.test_dir, 'baz.output')))

    # run tagged include which runs untagged include
    def test_run_tagged_untagged(self):
        self.populate_file('main.yaml', '''---
                include: 
                    file: one.yaml
                    as: one
                steps:
                    - run: 'one.before'
                ''')
        self.populate_file('one.yaml', '''---
                include: 
                    file: two.yaml
                    as: two
                steps:
                    - run:
                        include: two
                        tag: before
                    - echo: {from: '{{ bar }}', to: after.output, tag: after}
                ''')
        self.populate_file('two.yaml', '''---
                steps:
                    - echo: {from: '1', to: foo.output}
                    - echo: {from: '2', to: baz.output}
                    - echo: {from: '3', to: bar.output}
                        ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'foo.output'), '1'))
        self.assertTrue(check_file(join(self.test_dir, 'baz.output'), '2'))
        self.assertTrue(check_file(join(self.test_dir, 'bar.output'), '3'))

    # run tagged include which runs tagged include
    def test_run_tagged_tagged(self):
        self.populate_file('main.yaml', '''---
                        include: 
                            file: one.yaml
                            as: one
                        steps:
                            - run: 'one.before'
                        ''')
        self.populate_file('one.yaml', '''---
                        include: 
                            file: two.yaml
                            as: two
                        steps:
                            - run:
                                include: two.one
                                tag: before
                            - echo: {from: '{{ bar }}', to: after.output, tag: after}
                        ''')
        self.populate_file('two.yaml', '''---
                        steps:
                            - echo: {from: '1', to: foo.output, tag: one}
                            - echo: {from: '2', to: baz.output, tag: two}
                            - echo: {from: '3', to: bar.output, tag: three}
                                ''')
        runner = Runner(self.test_dir, join(self.test_dir, 'main.yaml'), None)
        runner.run_tests()
        self.assertTrue(check_file(join(self.test_dir, 'foo.output'), '1'))
        self.assertTrue(not os.path.exists(join(self.test_dir, 'baz.output')))
        self.assertTrue(not os.path.exists(join(self.test_dir, 'bar.output')))



import types
from abc import abstractmethod
from functools import wraps

from catcher.utils.misc import try_get_object, merge_two_dicts, fill_template

registered_steps = {}


def register_class(target_class):
    registered_steps[target_class.__name__.lower()] = target_class


class MetaStep(type):
    def __new__(mcs, name, bases, class_dict):
        cls = type.__new__(mcs, name, bases, class_dict)
        if cls.__name__ != 'Step' or cls.__name__ != 'ExternalStep':
            register_class(cls)
        return cls


SERVICE_KEYS = ['register', 'ignore_errors', 'name', 'tag', 'skip_if', 'run_if']


class SkipException(Exception):
    def __init__(self, message):
        super().__init__(message)


class Step(object, metaclass=MetaStep):
    """
    Abstract class for all Steps. Operates with common properties, available for all steps.

    :register: register a new variable at the end of this step. *Optional*
    :name: Set name for this step. Is used for passed steps output. *Optional*
    :ignore_errors: Do not stop the test if this step fails. Can be useful with running includes. *Optional*
    :tag: Tag this step to be called via `run` with tag. *Optional*
    :skip_if: Skip condition. This step will be skipped if condition is true. *Optional*
    :run_if: Run type for final action. *Optional*. 'pass' will run action only if test passes,
             'fail' will run action only if test fails. 'always' will always run action. It is the default value.
    :actions: Each step can have one ore multiple actions. In case of one action `actions` list is
                not necessary and you can use short form. Also - in case of several actions each should have its
                own properties like `register`, `tag` etc...

    One action - short form.
    ::

         http:
             post:  # register client and get id
                  url: '{{ user_service_url }}/sign_up'
                  body: {email: '{{ email }}', name: '{{ user }}'}

    Multiple actions.
    ::

        http:
            actions:
              - post:  # register client and get id
                  url: '{{ user_service_url }}/sign_up'
                  body: {email: '{{ email }}', name: 'TestUser'}
              - post:  # fill some personal data
                  url: '{{ user_service_url }}/data'
                  body: {gender: 'M', age: 22, firstName: 'John', lastName: 'Doe'}


    :Examples:

    Register new variable after `echo` step finishes.
    ::

        echo: {from: '{{ RANDOM_STR }}@test.com', register: {user_email: '{{ OUTPUT }}'}}

    Tag both http actions with `sign_up`
    ::

        http:
            actions:
              - post:  # register client and get id
                  url: '{{ user_service_url }}/sign_up'
                  headers: {Content-Type: 'application/json'}
                  body: {email: '{{ email }}', name: 'TestUser'}
                  response_code: 201
                register: {token: '{{ OUTPUT.data.token }}'}
                tag: sign_up
              - post:  # fill some personal data
                  url: '{{ user_service_url }}/data'
                  headers: {Content-Type: 'application/json', Authorization: '{{ token }}'}
                  body: {gender: 'M', age: 22, firstName: 'John', lastName: 'Doe'}
                register: {uuid: '{{ OUTPUT.data.uuid }}'}
                tag: sign_up

    Use custom name for the step
    ::

         http:
             post:  # register client and get id
                  url: '{{ user_service_url }}/sign_up'
                  headers: {Content-Type: 'application/json'}
                  body: {email: '{{ email }}', name: '{{ user }}'}
                  response_code: 201
             name: 'Register {{ user }} on remote server'

    Ignore errors and continue to another step
    ::

        http: {get: {url: 'http://test.com', response_code: 200}, ignore_errors: true}

    Skip one step based on variable got
    ::

        steps:
            - http:
                get:
                    url: '{{ my_web_service }}/api/v1/users?id={{ user_id }}'
                register: {registration_type: '{{ OUTPUT.data.registration }}'}
                name: 'Determine registration type for user {{ user_id }}'
            - postgres:
                request:
                    conf: 'test:test@localhost:5433/test'
                    query: "insert into loans(value) values(1000) where user_id == '{{ user_id }}';"
                name: 'Update user loan for facebook user'
                skip_if:
                    equals: {the: '{{ registration_type }}', is_not: 'facebook'}
            - couchbase:
                request:
                    conf:
                        bucket: loans
                        host: localhost
                    put:
                        key: '{{ user_id }}'
                        value: {value: 1000}
                skip_if:
                    equals: {the: '{{ registration_type }}', is_not: 'other'}

    Run step and do some clean up after
    ::

        steps:
            - http:
                get:
                    url: '{{ my_web_service }}/api/v1/users?id={{ user_id }}'
                register: {registration_type: '{{ OUTPUT.data.registration }}'}
                name: 'Determine registration type for user {{ user_id }}'
            - postgres:
                request:
                    conf: '{{ postgres_conf }}'
                    query: "insert into loans(value) values(1000) where user_id == '{{ user_id }}';"
                name: 'Update user loan for facebook user'
        finally:
            - postgres:
                request:
                    conf: '{{ postgres_conf }}'
                    query: "delete from loans(value) where user_id == '{{ user_id }}';"
                name: 'Clean up user'

    """

    def __init__(self, register=None, name=None, ignore_errors=False, skip_if=None, **kwargs) -> None:
        self.register = register
        self.name = name
        self.ignore_errors = ignore_errors
        self.skip_if = skip_if

    @abstractmethod
    def action(self, includes: dict, variables: dict) -> dict or tuple:
        """
        Perform an action.

        :param includes: Script includes.
        :param variables: Script variables.
        :return: step output

        For code above
        ::

            postgres:
                request:
                    conf:
                        dbname: test
                        user: test
                        password: test
                        host: localhost
                        port: 5433
                    query: 'select count(*) from test'
            register: {documents: '{{ OUTPUT.count }}'}

        step's input will be
        ::

            {'request' : {'conf': {'dbname': 'test', 'user': 'test', 'password': 'test', 'host': 'localhost', 'port': 5433},
                          'query': 'select count(*) from test'}
            }

        """
        pass

    def check_skip(self, variables: dict):
        if self.skip_if is None:
            return False
        from catcher.steps.check import Operator
        operator = Operator.find_operator(self.skip_if)
        if operator.operation(variables):
            raise SkipException('Skipped due to {}'.format(self.skip_if))

    @staticmethod
    def filter_predefined_keys(kwargs: dict):
        [action] = [k for k in kwargs.keys() if k not in SERVICE_KEYS and not k.startswith('_')]
        return action

    def process_register(self, variables, output: dict or list or str or None = None) -> dict:
        if self.register is not None:
            for key in self.register.keys():
                if output is not None:
                    out = fill_template(self.register[key],
                                        merge_two_dicts(variables, {'OUTPUT': try_get_object(output)}))
                else:
                    out = fill_template(self.register[key], variables)
                variables[key] = out
        return variables


def update_variables(func):
    """
    Use this decorator on Step.action implementation.

    Your action method should always return variables, or
    both variables and output.

    This decorator will update variables with output.

    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        if isinstance(result, tuple):
            return self.process_register(result[0], result[1])
        else:
            return self.process_register(result)

    return wrapper



from abc import ABC, abstractmethod
from os import listdir
from os.path import join
from typing import Union


class Module(ABC):

    def __init__(self, resources) -> None:
        super().__init__()
        self._resources = resources

    @abstractmethod
    def before(self, *args, **kwargs):
        """
        Run before tests execution
        """
        pass

    @abstractmethod
    def after(self, *args, **kwargs):
        """
        Run after tests execution
        """
        pass

    def find_resource_file(self) -> Union[str, None]:
        """ find a file inside a resources directory """
        if self._resources is None:
            return None
        try:
            files = [f for f in listdir(self._resources) if self.filter_resource(f)]
            if files:
                return join(self._resources, files[0])
            return None
        except FileNotFoundError:
            return None

    @abstractmethod
    def filter_resource(self, file: str) -> bool:
        pass



from catcher.steps.step import Step, update_variables, SkipException
from catcher.utils.logger import error, info, debug
from catcher.utils.misc import merge_two_dicts, fill_template_str, fill_template, fill_template_recursive
from catcher.steps.stop import StopException
from catcher.utils import logger


class Run(Step):
    """
    Run another test, included later. Is useful when you need to run the same code from different tests or to
    repeat the same steps inside one test but with different input variables.

    :Input:

    :include: include name. If contains dot - everything after dot will be considered as tag. In case of multiple dots
      the last one will be considered as tag.
    :variables: Variables to override. *Optional*

    :Examples:

    Use short form to run `sign_up`
    ::

        include:
            file: register_user.yaml
            as: sign_up
        steps:
            # .... some steps
            - run: sign_up
            # .... some steps

    Run `sign_up` include test twice for 2 different users
    ::

        include:
            file: register_user.yaml
            as: sign_up
        variables:
            users: ['{{ random("email") }}', '{{ random("email") }}']
        steps:
            # .... some steps
            - run:
                include: sign_up
                variables:
                    user: '{{ users[0] }}'
            # .... some steps
            - run:
                include: sign_up
                variables:
                    user: '{{ users[1] }}'

    Include `sign_up` and run all steps with tag `register` from it.
    ::

        include:
            file: register_and_login.yaml
            as: sign_up
        steps:
            - run:
                include: sign_up.register

    Include one.yaml from main and run only **before** tag of it. **main.yaml** -> **one.yaml** -> **two.yaml**.
    main.yaml ::

        include:
            file: one.yaml
            as: one
        steps:
            - run: 'one.before'

    one.yaml ::

        include:
            file: two.yaml
            as: run_me
        steps:
            - run:
                include: two.run_me
                tag: before
            - echo: {from: '{{ bar }}', to: after.output, tag: after}

    two.yaml ::

        steps:
            - echo: {from: '1', to: foo.output, tag: run_me}
            - echo: {from: '2', to: baz.output, tag: two}
            - echo: {from: '3', to: bar.output, tag: three}
    """

    def __init__(self, ignore_errors=False, _body=None, run=None, include=None, tag=None, variables=None, **kwargs):
        super().__init__(**kwargs)
        self.variables = {} if variables is None else variables
        self.ignore_errors = ignore_errors
        self.include = include
        self.tag = tag
        if self.include is None:
            if run and isinstance(run, str):
                self.include = run
            else:
                self.include = _body

    @update_variables
    def action(self, includes: dict, variables: dict) -> dict:
        filled_vars = dict([(k, fill_template(v, variables)) for (k, v) in self.variables.items()])
        out = fill_template_str(self.include, variables)
        test, tag = get_tag(out)
        if test not in includes:
            error('No include registered for name ' + test)
            raise Exception('No include registered for name ' + test)
        include = includes[test]
        variables = merge_two_dicts(include.variables, merge_two_dicts(variables, filled_vars))
        include.variables = fill_template_recursive(variables, variables)
        try:
            info('Running {}.{}'.format(test, '' if tag is None else tag))
            logger.log_storage.nested_test_in()
            variables = include.run(tag=tag, raise_stop=True)
            logger.log_storage.nested_test_out()
        except SkipException:
            logger.log_storage.nested_test_out()
            debug('Include ignored')
            return variables
        except StopException as e:
            logger.log_storage.nested_test_out()
            raise e
        except Exception as e:
            logger.log_storage.nested_test_out()
            if not self.ignore_errors:
                raise Exception('Step run ' + test + ' failed: ' + str(e))
        return variables


def get_tag(include: str) -> str or None:
    if '.' in include:
        splitted = include.split('.')
        return '.'.join(include.split('.')[:-1]), splitted[-1:][0]
    else:
        return include, None
